# -*- coding: utf-8 -*-
"""
Expression for unit cell parameters are presented.
For details see documentation.
"""
import warnings
import numpy 

from .matrix_operations import calc_m_q_mt


def calc_reciprocal_by_unit_cell_parameters(
        unit_cell_parameters, flag_unit_cell_parameters: bool = False):
    a, b = unit_cell_parameters[0], unit_cell_parameters[1]
    c = unit_cell_parameters[2]

    alpha, beta = unit_cell_parameters[3], unit_cell_parameters[4]
    gamma = unit_cell_parameters[5]
    c_a, c_b, c_g = numpy.cos(alpha), numpy.cos(beta), numpy.cos(gamma)
    s_a, s_b, s_g = numpy.sin(alpha), numpy.sin(beta), numpy.sin(gamma)
    vol, dder_vol = \
        calc_volume_uc_by_unit_cell_parameters(
            unit_cell_parameters, flag_unit_cell_parameters=flag_unit_cell_parameters)

    ialpha = numpy.arccos((c_b*c_g-c_a)/(s_b*s_g))
    ibeta = numpy.arccos((c_g*c_a-c_b)/(s_g*s_a))
    igamma = numpy.arccos((c_a*c_b-c_g)/(s_a*s_b))
    ia, ib, ic = b*c*s_a/vol, c*a*s_b/vol, a*b*s_g/vol
    reciprocal_unit_cell_parameters = numpy.stack([
        ia, ib, ic, ialpha, ibeta, igamma], axis=0)
    
    dder = {}
    if flag_unit_cell_parameters:
        d_v = dder_vol["unit_cell_parameters"]
        zeros = numpy.zeros(unit_cell_parameters[0].shape, dtype=float)
        i_vol_sq = 1./numpy.square(vol)
        der_ia = numpy.stack([b*c*s_a*d_v[0]*i_vol_sq,
                              c*s_a/vol+b*c*s_a*d_v[1]*i_vol_sq,
                              b*s_a/vol+b*c*s_a*d_v[2]*i_vol_sq,
                              b*c*c_a/vol+b*c*s_a*d_v[3]*i_vol_sq,
                              b*c*s_a*d_v[4]*i_vol_sq,
                              b*c*s_a*d_v[5]*i_vol_sq], axis=0)
        der_ib = numpy.stack([c*s_b/vol+c*a*s_b*i_vol_sq*d_v[0],
                              c*a*s_b*i_vol_sq*d_v[1],
                              a*s_b/vol+c*a*s_b*i_vol_sq*d_v[2],
                              c*a*s_b*i_vol_sq*d_v[3],
                              c*a*c_b/vol+c*a*s_b*i_vol_sq*d_v[4],
                              c*a*s_b*i_vol_sq*d_v[5]], axis=0)
        der_ic = numpy.stack([zeros, zeros, zeros, zeros, zeros, zeros], axis=0)
        der_ial = numpy.stack([zeros, zeros, zeros, zeros, zeros, zeros], axis=0)
        der_ibe = numpy.stack([zeros, zeros, zeros, zeros, zeros, zeros], axis=0)
        der_iga = numpy.stack([zeros, zeros, zeros, zeros, zeros, zeros], axis=0)

        dder["unit_cell_parameters"] = numpy.stack([
            der_ia, der_ib, der_ic, der_ial, der_ibe, der_iga], axis=0)
    return reciprocal_unit_cell_parameters, dder


def calc_phi_sq_by_unit_cell_parameters(
        unit_cell_parameters, flag_unit_cell_parameters: bool = False):
    """
    Calculate phi_sq
    phi_sq = 1 - cos^2 alpha - cos^2 beta - cos^2 gamma - 2 cos alpha cos beta cos gamma
    """
    alpha, beta = unit_cell_parameters[3], unit_cell_parameters[4]
    gamma = unit_cell_parameters[5]
    c_a, c_b, c_g = numpy.cos(alpha), numpy.cos(beta), numpy.cos(gamma)
    phi_sq =  (
        1. - numpy.square(c_a) - numpy.square(c_b) - numpy.square(c_g) +
        2.*c_a*c_b*c_g)
    dder = {}
    if flag_unit_cell_parameters:
        der_a = numpy.zeros_like(unit_cell_parameters[0])
        s_a, s_b, s_g = numpy.sin(alpha), numpy.sin(beta), numpy.sin(gamma)
        der_al = 2*s_a*(c_a-c_b*c_g)
        der_be = 2*s_b*(c_b-c_g*c_a)
        der_ga = 2*s_g*(c_g-c_a*c_b)
        dder["unit_cell_parameters"] = numpy.stack([
            der_a, der_a, der_a, der_al, der_be, der_ga], axis=0)
    return phi_sq, dder


def calc_phi_by_unit_cell_parameters(
        unit_cell_parameters, flag_unit_cell_parameters: bool = False):
    """
    Calculate phi
    phi = (1 - cos^2 alpha - cos^2 beta - cos^2 gamma - 2 cos alpha cos beta cos gamma)^0.5
    """
    phi_sq, dder_phi_sq = calc_phi_sq_by_unit_cell_parameters(
        unit_cell_parameters, flag_unit_cell_parameters)
    phi = numpy.sqrt(phi_sq)
    dder = {}
    if flag_unit_cell_parameters:
        dder["unit_cell_parameters"] = 0.5*dder_phi_sq["unit_cell_parameters"]/phi 
    return phi, dder


def calc_volume_uc_by_unit_cell_parameters(
        unit_cell_parameters, flag_unit_cell_parameters: bool = False):
    """
    Calculate the volume of unit cell defined as 
    a, b, c, alpha, beta, gamma.
    Angles should be given in radians.
    """
    a, b = unit_cell_parameters[0], unit_cell_parameters[1]
    c = unit_cell_parameters[2]
    alpha, beta = unit_cell_parameters[3], unit_cell_parameters[4]
    gamma = unit_cell_parameters[5]
    c_a, c_b, c_g = numpy.cos(alpha), numpy.cos(beta), numpy.cos(gamma)
    c_a_sq, c_b_sq = numpy.square(c_a), numpy.square(c_b)
    c_g_sq = numpy.square(c_g)
    phi, dder_phi = calc_phi_by_unit_cell_parameters(unit_cell_parameters, flag_unit_cell_parameters)
    volume_uc = a*b*c*phi
    dder = {}
    if flag_unit_cell_parameters:
        s_a, s_b, s_g = numpy.sin(alpha), numpy.sin(beta), numpy.sin(gamma)
        der_a = b*c*phi
        der_b = a*c*phi
        der_c = a*b*phi
        der_al = -a*b*c*0.5*(-2*c_a_sq+2.*c_b*c_g)*s_a/phi
        der_be = -a*b*c*0.5*(-2*c_b_sq+2.*c_a*c_g)*s_b/phi
        der_ga = -a*b*c*0.5*(-2*c_g_sq+2.*c_a*c_b)*s_g/phi
        dder["unit_cell_parameters"] = numpy.stack([
            der_a, der_b, der_c, der_al, der_be, der_ga], axis=0)
    return volume_uc, dder


def calc_volume_ruc_by_unit_cell_parameters(
        unit_cell_parameters, flag_unit_cell_parameters: bool = False):
    """
    Calculate the volume of reciprocal unit cell 
    """
    volume_uc , dder_volume_uc = \
        calc_volume_uc_by_unit_cell_parameters(
            unit_cell_parameters,
            flag_unit_cell_parameters=flag_unit_cell_parameters)
    volume_ruc = 1./volume_uc
    dder = {}
    if flag_unit_cell_parameters:
        dder["unit_cell_parameters"] = -numpy.square(volume_ruc)*dder_volume_uc["unit_cell_parameters"]
    return volume_ruc, dder


def calc_m_g_by_unit_cell_parameters(
        unit_cell_parameters, flag_unit_cell_parameters: bool = False):
    a, b = unit_cell_parameters[0], unit_cell_parameters[1]
    c = unit_cell_parameters[2]
    alpha, beta = unit_cell_parameters[3], unit_cell_parameters[4]
    gamma = unit_cell_parameters[5]
    c_a, c_b, c_g = numpy.cos(alpha), numpy.cos(beta), numpy.cos(gamma)

    m_g_ij = numpy.stack([a*a,b*b, c*c, a*b*c_g, a*c*c_b,b*c*c_a])
    dder = {}
    if flag_unit_cell_parameters:
        s_a, s_b, s_g = numpy.sin(alpha), numpy.sin(beta), numpy.sin(gamma)
        zeros = numpy.zeros_like(unit_cell_parameters[0])
        der_11 = numpy.stack([2*a, zeros, zeros, zeros, zeros, zeros], axis=0)
        der_12 = numpy.stack([b*c_g, a*c_g, zeros, zeros, zeros, -a*b*s_g],
                             axis=0)
        der_13 = numpy.stack([c*c_b, zeros, a*c_b, zeros, -a*c*s_b, zeros],
                             axis=0)
        der_22 = numpy.stack([zeros, 2*b, zeros, zeros, zeros, zeros], axis=0)
        der_23 = numpy.stack([zeros, c*c_a, b*c_a, -b*c*s_a, zeros, zeros], axis=0)
        der_33 = numpy.stack([zeros, zeros, 2*c, zeros, zeros, zeros], axis=0)

        dder["unit_cell_parameters"] = numpy.stack([
            der_11, der_22, der_33, der_12, der_13, der_23], axis=0)
    return m_g_ij, dder


def calc_m_g_norm_by_unit_cell_parameters(
        unit_cell_parameters, flag_unit_cell_parameters: bool = False):
    alpha, beta = unit_cell_parameters[3], unit_cell_parameters[4]
    gamma = unit_cell_parameters[5]
    c_a, c_b, c_g = numpy.cos(alpha), numpy.cos(beta), numpy.cos(gamma)

    ones = numpy.ones_like(c_a)

    m_g_norm_ij = numpy.stack([ones,ones,ones, c_g, c_b, c_a])
    dder = {}
    if flag_unit_cell_parameters:
        zeros = numpy.zeros_like(c_a)

        s_a, s_b, s_g = numpy.sin(alpha), numpy.sin(beta), numpy.sin(gamma)
        der_11 = numpy.stack([zeros, zeros, zeros, zeros, zeros, zeros], axis=0)
        der_12 = numpy.stack([c_g, c_g, zeros, zeros, zeros, -s_g], axis=0)
        der_13 = numpy.stack([c_b, zeros, c_b, zeros, -s_b, zeros], axis=0)
        der_22 = numpy.stack([zeros, zeros, zeros, zeros, zeros, zeros], axis=0)
        der_23 = numpy.stack([zeros, c_a, c_a, -s_a, zeros, zeros], axis=0)
        der_33 = numpy.stack([zeros, zeros, zeros, zeros, zeros, zeros], axis=0)

        dder["unit_cell_parameters"] = numpy.stack([
            der_11, der_22, der_33, der_12, der_13, der_23], axis=0)
    return m_g_norm_ij, dder


def calc_m_reciprocal_g_by_unit_cell_parameters(
        unit_cell_parameters, flag_unit_cell_parameters: bool = False):
    a, b = unit_cell_parameters[0], unit_cell_parameters[1]
    c = unit_cell_parameters[2]
    alpha, beta = unit_cell_parameters[3], unit_cell_parameters[4]
    gamma = unit_cell_parameters[5]
    c_a, c_b, c_g = numpy.cos(alpha), numpy.cos(beta), numpy.cos(gamma)
    s_a, s_b, s_g = numpy.sin(alpha), numpy.sin(beta), numpy.sin(gamma)
    phi_sq, dder_phi_sq = calc_phi_sq_by_unit_cell_parameters(
        unit_cell_parameters, flag_unit_cell_parameters=flag_unit_cell_parameters) 

    inv_g_11 = numpy.square(s_a) / (numpy.square(a)* phi_sq)
    inv_g_22 = numpy.square(s_b) / (numpy.square(b)* phi_sq)
    inv_g_33 = numpy.square(s_g) / (numpy.square(c)* phi_sq)
    inv_g_12 = (c_a*c_b-c_g) / (a*b* phi_sq)
    inv_g_13 = (c_a*c_g-c_b) / (a*c* phi_sq)
    inv_g_23 = (c_b*c_g-c_a) / (b*c* phi_sq)

    # G matrix for reciprocal unit cell
    m_reciprocal_g_ij = numpy.stack(
        [inv_g_11, inv_g_22, inv_g_33,
         inv_g_12, inv_g_13, inv_g_23],
        axis=0)
    dder = {}
    if flag_unit_cell_parameters:
        zeros = numpy.zeros_like(unit_cell_parameters[0])
        der_11 = numpy.stack([zeros, zeros, zeros, zeros, zeros, zeros], axis=0)
        der_12 = numpy.stack([zeros, zeros, zeros, zeros, zeros, zeros], axis=0)
        der_13 = numpy.stack([zeros, zeros, zeros, zeros, zeros, zeros], axis=0)
        der_22 = numpy.stack([zeros, zeros, zeros, zeros, zeros, zeros], axis=0)
        der_23 = numpy.stack([zeros, zeros, zeros, zeros, zeros, zeros], axis=0)
        der_33 = numpy.stack([zeros, zeros, zeros, zeros, zeros, zeros], axis=0)

        dder["unit_cell_parameters"] = numpy.stack([
            der_11, der_22, der_33, der_12, der_13, der_23], axis=0)
    return m_reciprocal_g_ij, dder


def calc_m_reciprocal_g_norm_by_unit_cell_parameters(
        unit_cell_parameters, flag_unit_cell_parameters: bool = False):
    a, b = unit_cell_parameters[0], unit_cell_parameters[1]
    c = unit_cell_parameters[2]
    alpha, beta = unit_cell_parameters[3], unit_cell_parameters[4]
    gamma = unit_cell_parameters[5]
    c_a, c_b, c_g = numpy.cos(alpha), numpy.cos(beta), numpy.cos(gamma)
    s_a, s_b, s_g = numpy.sin(alpha), numpy.sin(beta), numpy.sin(gamma)
    phi_sq, dder_phi_sq = calc_phi_sq_by_unit_cell_parameters(
        unit_cell_parameters, flag_unit_cell_parameters=flag_unit_cell_parameters) 

    ones = numpy.ones_like(c_a)

    inv_g_norm_11 = ones
    inv_g_norm_22 = ones
    inv_g_norm_33 = ones
    inv_g_norm_12 = (c_a*c_b-c_g) / (s_a*s_b)
    inv_g_norm_13 = (c_a*c_g-c_b) / (s_a*s_g)
    inv_g_norm_23 = (c_b*c_g-c_a) / (s_b*s_g)

    # G matrix for reciprocal unit cell
    m_reciprocal_g_norm_ij = numpy.stack(
        [inv_g_norm_11, inv_g_norm_22, inv_g_norm_33,
         inv_g_norm_12, inv_g_norm_13, inv_g_norm_23],
        axis=0)
    dder = {}
    if flag_unit_cell_parameters:
        zero = numpy.zeros_like(a)
        s_a_sq = numpy.square(s_a)
        s_b_sq = numpy.square(s_b)
        s_g_sq = numpy.square(s_g)

        der_11 = numpy.stack([zero, zero, zero, zero, zero, zero], axis=0)
        der_22 = numpy.stack([zero, zero, zero, zero, zero, zero], axis=0)
        der_33 = numpy.stack([zero, zero, zero, zero, zero, zero], axis=0)
        der_12 = numpy.stack([zero, zero, zero, -(c_a*c_b - c_g)*c_a/(s_a_sq*s_b) - c_b/s_b, -(c_a*c_b - c_g)*c_b/(s_a*s_b_sq) - c_a/s_a, s_g/(s_a*s_b)], axis=0)
        der_13 = numpy.stack([zero, zero, zero, -(c_a*c_g - c_b)*c_a/(s_a_sq*s_g) - c_g/s_g, s_b/(s_a*s_g), -(c_a*c_g - c_b)*c_g/(s_a*s_g_sq) - c_a/s_a], axis=0)
        der_23 = numpy.stack([zero, zero, zero, s_a/(s_b*s_g), -(-c_a + c_b*c_g)*c_b/(s_b_sq*s_g) - c_g/s_g, -(-c_a + c_b*c_g)*c_g/(s_b*s_g_sq) - c_b/s_b], axis=0)

        dder["unit_cell_parameters"] = numpy.stack([
            der_11, der_22, der_33, der_12, der_13, der_23], axis=0)
    return m_reciprocal_g_norm_ij, dder


def calc_m_b_by_unit_cell_parameters(
        unit_cell_parameters, flag_unit_cell_parameters: bool = False):
    a, b = unit_cell_parameters[0], unit_cell_parameters[1]
    c = unit_cell_parameters[2]
    alpha, beta = unit_cell_parameters[3], unit_cell_parameters[4]
    gamma = unit_cell_parameters[5]
    c_a, c_b, c_g = numpy.cos(alpha), numpy.cos(beta), numpy.cos(gamma)
    s_a, s_b, s_g = numpy.sin(alpha), numpy.sin(beta), numpy.sin(gamma)

    phi, dder_phi = calc_phi_by_unit_cell_parameters(
        unit_cell_parameters, flag_unit_cell_parameters=flag_unit_cell_parameters) 

    zeros = numpy.zeros_like(unit_cell_parameters[0])

    m_b_ij = numpy.stack([
        s_a/(a*phi), (c_a*c_b-c_g)/(b*phi*s_a), (c_a*c_g-c_b)/(c*phi*s_a),
              zeros, 1/(b*s_a), -c_a/(c*s_a),
              zeros,     zeros, 1./c], axis=0)
    dder = {}
    if flag_unit_cell_parameters:
        der_11_a = - m_b_ij[0]/a - m_b_ij[0]/phi*dder_phi["unit_cell_parameters"][0]
        der_11_b = - m_b_ij[0]/phi*dder_phi["unit_cell_parameters"][1]
        der_11_c = - m_b_ij[0]/phi*dder_phi["unit_cell_parameters"][2]
        der_11_al = c_a/(a*phi) - m_b_ij[0]/phi*dder_phi["unit_cell_parameters"][3]
        der_11_be = - m_b_ij[0]/phi*dder_phi["unit_cell_parameters"][4]
        der_11_ga = - m_b_ij[0]/phi*dder_phi["unit_cell_parameters"][5]
        der_11 = numpy.stack([der_11_a, der_11_b, der_11_c, der_11_al, der_11_be, der_11_ga], axis=0)

        der_12_a = - m_b_ij[1]/phi*dder_phi["unit_cell_parameters"][0]
        der_12_b = - m_b_ij[1]/b - m_b_ij[1]/phi*dder_phi["unit_cell_parameters"][1]
        der_12_c = - m_b_ij[1]/phi*dder_phi["unit_cell_parameters"][2]
        der_12_al = (c_a*c_g-c_b)/(numpy.square(s_a)*b*phi) - m_b_ij[1]/phi*dder_phi["unit_cell_parameters"][3]
        der_12_be = c_a*s_b/(b*phi*s_a) - m_b_ij[1]/phi*dder_phi["unit_cell_parameters"][4]
        der_12_ga = s_g/(b*phi*s_a) - m_b_ij[1]/phi*dder_phi["unit_cell_parameters"][5]
        der_12 = numpy.stack([der_12_a, der_12_b, der_12_c, der_12_al, der_12_be, der_12_ga], axis=0)

        der_13_a = - m_b_ij[2]/phi*dder_phi["unit_cell_parameters"][0]
        der_13_b = - m_b_ij[2]/phi*dder_phi["unit_cell_parameters"][1]
        der_13_c = -m_b_ij[2]/c - m_b_ij[2]/phi*dder_phi["unit_cell_parameters"][2]
        der_13_al = (c_a*c_b-c_g)/(numpy.square(s_a)*c*phi) - m_b_ij[2]/phi*dder_phi["unit_cell_parameters"][3]
        der_13_be = s_b/(c*phi*s_a) - m_b_ij[2]/phi*dder_phi["unit_cell_parameters"][4]
        der_13_ga = -s_g*c_a/(c*phi*s_a) - m_b_ij[2]/phi*dder_phi["unit_cell_parameters"][5]
        der_13 = numpy.stack([der_13_a, der_13_b, der_13_c, der_13_al, der_13_be, der_13_ga], axis=0)

        der_22_b =  -m_b_ij[4]/b
        der_22_al = -m_b_ij[4]*c_a/numpy.square(s_a)
        der_22 = numpy.stack([zeros, der_22_b, zeros, der_22_al, zeros, zeros], axis=0)

        der_23_c =  -m_b_ij[5]/c
        der_23_al = 1/(numpy.square(s_a)*c)
        der_23 = numpy.stack([zeros, zeros, der_23_c, der_23_al, zeros, zeros], axis=0)

        der_33_c =  -1/numpy.square(c)
        der_33 = numpy.stack([zeros, zeros, der_33_c, zeros, zeros, zeros], axis=0)

        der_21 = numpy.stack([zeros, zeros, zeros, zeros, zeros, zeros], axis=0)
        der_31 = numpy.stack([zeros, zeros, zeros, zeros, zeros, zeros], axis=0)
        der_32 = numpy.stack([zeros, zeros, zeros, zeros, zeros, zeros], axis=0)

        dder["unit_cell_parameters"] = numpy.stack([
            der_11, der_12, der_13, der_21, der_22, der_23, der_31, der_32,
            der_33], axis=0)
    return m_b_ij, dder


def calc_m_inv_b_by_unit_cell_parameters(
        unit_cell_parameters, flag_unit_cell_parameters: bool = False):
    a, b = unit_cell_parameters[0], unit_cell_parameters[1]
    c = unit_cell_parameters[2]
    alpha, beta = unit_cell_parameters[3], unit_cell_parameters[4]
    gamma = unit_cell_parameters[5]
    c_a, c_b, c_g = numpy.cos(alpha), numpy.cos(beta), numpy.cos(gamma)
    s_a, s_b, s_g = numpy.sin(alpha), numpy.sin(beta), numpy.sin(gamma)

    phi, dder_phi = calc_phi_by_unit_cell_parameters(
        unit_cell_parameters, flag_unit_cell_parameters=flag_unit_cell_parameters) 

    zeros = numpy.zeros_like(unit_cell_parameters[0])

    m_inv_b_ij = numpy.stack([
        (a*phi)/s_a, a*(c_g-c_a*c_b)/s_a, a*c_b,
              zeros,               b*s_a, b*c_a,
              zeros,               zeros, c], axis=0)
    dder = {}
    if flag_unit_cell_parameters:
        der_11 = numpy.stack([zeros, zeros, zeros, zeros, zeros, zeros], axis=0)
        der_12 = numpy.stack([zeros, zeros, zeros, zeros, zeros, zeros], axis=0)
        der_13 = numpy.stack([zeros, zeros, zeros, zeros, zeros, zeros], axis=0)
        der_21 = numpy.stack([zeros, zeros, zeros, zeros, zeros, zeros], axis=0)
        der_22 = numpy.stack([zeros, zeros, zeros, zeros, zeros, zeros], axis=0)
        der_23 = numpy.stack([zeros, zeros, zeros, zeros, zeros, zeros], axis=0)
        der_31 = numpy.stack([zeros, zeros, zeros, zeros, zeros, zeros], axis=0)
        der_32 = numpy.stack([zeros, zeros, zeros, zeros, zeros, zeros], axis=0)
        der_33 = numpy.stack([zeros, zeros, zeros, zeros, zeros, zeros], axis=0)

        dder["unit_cell_parameters"] = numpy.stack([
            der_11, der_12, der_13, der_21, der_22, der_23, der_31, der_32,
            der_33], axis=0)
    return m_inv_b_ij, dder


def calc_m_b_norm_by_unit_cell_parameters(
        unit_cell_parameters, flag_unit_cell_parameters: bool = False):
    a, b = unit_cell_parameters[0], unit_cell_parameters[1]
    c = unit_cell_parameters[2]
    alpha, beta = unit_cell_parameters[3], unit_cell_parameters[4]
    gamma = unit_cell_parameters[5]
    c_a, c_b, c_g = numpy.cos(alpha), numpy.cos(beta), numpy.cos(gamma)
    s_a, s_b, s_g = numpy.sin(alpha), numpy.sin(beta), numpy.sin(gamma)

    phi, dder_phi = calc_phi_by_unit_cell_parameters(
        unit_cell_parameters, flag_unit_cell_parameters=flag_unit_cell_parameters) 

    zeros = numpy.zeros_like(unit_cell_parameters[0])
    ones = numpy.ones_like(unit_cell_parameters[0])


    m_b_norm_ij = numpy.stack(
        [ones, (c_a*c_b-c_g)/(s_a*s_b), (c_a*c_g-c_b)/(s_a*s_g),
         zeros, phi/(s_a*s_b), -phi*c_a/(s_a*s_g),
         zeros, zeros, phi/s_g], axis=0)
 
    dder = {}
    if flag_unit_cell_parameters:
        der_11 = numpy.stack([zeros, zeros, zeros, zeros, zeros, zeros], axis=0)
        der_12 = numpy.stack([zeros, zeros, zeros, zeros, zeros, zeros], axis=0)
        der_13 = numpy.stack([zeros, zeros, zeros, zeros, zeros, zeros], axis=0)
        der_21 = numpy.stack([zeros, zeros, zeros, zeros, zeros, zeros], axis=0)
        der_22 = numpy.stack([zeros, zeros, zeros, zeros, zeros, zeros], axis=0)
        der_23 = numpy.stack([zeros, zeros, zeros, zeros, zeros, zeros], axis=0)
        der_31 = numpy.stack([zeros, zeros, zeros, zeros, zeros, zeros], axis=0)
        der_32 = numpy.stack([zeros, zeros, zeros, zeros, zeros, zeros], axis=0)
        der_33 = numpy.stack([zeros, zeros, zeros, zeros, zeros, zeros], axis=0)

        dder["unit_cell_parameters"] = numpy.stack([
            der_11, der_12, der_13, der_21, der_22, der_23, der_31, der_32,
            der_33], axis=0)
    return m_b_norm_ij, dder


def calc_m_inv_b_norm_by_unit_cell_parameters(
        unit_cell_parameters, flag_unit_cell_parameters: bool = False):
    a, b = unit_cell_parameters[0], unit_cell_parameters[1]
    c = unit_cell_parameters[2]
    alpha, beta = unit_cell_parameters[3], unit_cell_parameters[4]
    gamma = unit_cell_parameters[5]
    c_a, c_b, c_g = numpy.cos(alpha), numpy.cos(beta), numpy.cos(gamma)
    s_a, s_b, s_g = numpy.sin(alpha), numpy.sin(beta), numpy.sin(gamma)

    phi, dder_phi = calc_phi_by_unit_cell_parameters(
        unit_cell_parameters, flag_unit_cell_parameters=flag_unit_cell_parameters) 

    zeros = numpy.zeros_like(unit_cell_parameters[0])
    ones = numpy.ones_like(unit_cell_parameters[0])


    m_inv_b_norm = numpy.stack(
        [ones, (c_g-c_a*c_b)/phi, (s_a*c_b)/phi,
         zeros, (s_a*s_b)/phi, (c_a*s_b)/phi,
         zeros, zeros, s_g/phi], axis=0)
 
    dder = {}
    if flag_unit_cell_parameters:

        der_11_1 = numpy.stack([zeros, zeros, zeros, zeros, zeros, zeros], axis=0)
        der_12_1 = numpy.stack([zeros, zeros, zeros, s_a*c_b/phi, s_b*c_a/phi, -s_g/phi], axis=0)
        der_13_1 = numpy.stack([zeros, zeros, zeros, c_a*c_b/phi, -s_a*s_b/phi, zeros], axis=0)
        der_22_1 = numpy.stack([zeros, zeros, zeros, s_b*c_a/phi, s_a*c_b/phi, zeros], axis=0)
        der_23_1 = numpy.stack([zeros, zeros, zeros, -s_a*s_b/phi, c_a*c_b/phi, zeros], axis=0)
        der_33_1 = numpy.stack([zeros, zeros, zeros, zeros, zeros, c_g/phi], axis=0)
        der_21_1 = numpy.stack([zeros, zeros, zeros, zeros, zeros, zeros], axis=0)
        der_31_1 = numpy.stack([zeros, zeros, zeros, zeros, zeros, zeros], axis=0)
        der_32_1 = numpy.stack([zeros, zeros, zeros, zeros, zeros, zeros], axis=0)

        dder_2 = -numpy.expand_dims(m_inv_b_norm/phi, axis=1)*numpy.expand_dims(dder_phi["uni_cell_parameters"], axis=0)

        dder["unit_cell_parameters"] = numpy.stack([
            der_11_1, der_12_1, der_13_1, der_21_1, der_22_1, der_23_1, der_31_1, der_32_1,
            der_33_1], axis=0) + dder_2
    return m_inv_b_norm, dder


def calc_m_m_by_unit_cell_parameters(
        unit_cell_parameters, flag_unit_cell_parameters: bool = False):
    a, b = unit_cell_parameters[0], unit_cell_parameters[1]
    c = unit_cell_parameters[2]
    alpha, beta = unit_cell_parameters[3], unit_cell_parameters[4]
    gamma = unit_cell_parameters[5]
    c_a, c_b, c_g = numpy.cos(alpha), numpy.cos(beta), numpy.cos(gamma)
    s_a, s_b, s_g = numpy.sin(alpha), numpy.sin(beta), numpy.sin(gamma)

    phi, dder_phi = calc_phi_by_unit_cell_parameters(
        unit_cell_parameters, flag_unit_cell_parameters=flag_unit_cell_parameters) 

    zeros = numpy.zeros_like(unit_cell_parameters[0])

    # M matrix
    m_m_ij = numpy.stack([
        a*phi/s_a, zeros, zeros,
        a*(c_g-c_a*c_b)/s_a, b*s_a, zeros,
        a*c_b, b*c_a, c], axis=0)
    dder = {}
    if flag_unit_cell_parameters:
        der_11 = numpy.stack([zeros, zeros, zeros, zeros, zeros, zeros], axis=0)
        der_12 = numpy.stack([zeros, zeros, zeros, zeros, zeros, zeros], axis=0)
        der_13 = numpy.stack([zeros, zeros, zeros, zeros, zeros, zeros], axis=0)
        der_21 = numpy.stack([zeros, zeros, zeros, zeros, zeros, zeros], axis=0)
        der_22 = numpy.stack([zeros, zeros, zeros, zeros, zeros, zeros], axis=0)
        der_23 = numpy.stack([zeros, zeros, zeros, zeros, zeros, zeros], axis=0)
        der_31 = numpy.stack([zeros, zeros, zeros, zeros, zeros, zeros], axis=0)
        der_32 = numpy.stack([zeros, zeros, zeros, zeros, zeros, zeros], axis=0)
        der_33 = numpy.stack([zeros, zeros, zeros, zeros, zeros, zeros], axis=0)

        dder["unit_cell_parameters"] = numpy.stack([
            der_11, der_12, der_13, der_21, der_22, der_23, der_31, der_32,
            der_33], axis=0)
    return m_m_ij, dder


def calc_m_inv_m_by_unit_cell_parameters(
        unit_cell_parameters, flag_unit_cell_parameters: bool = False):
    a, b = unit_cell_parameters[0], unit_cell_parameters[1]
    c = unit_cell_parameters[2]
    alpha, beta = unit_cell_parameters[3], unit_cell_parameters[4]
    gamma = unit_cell_parameters[5]
    c_a, c_b, c_g = numpy.cos(alpha), numpy.cos(beta), numpy.cos(gamma)
    s_a, s_b, s_g = numpy.sin(alpha), numpy.sin(beta), numpy.sin(gamma)

    phi, dder_phi = calc_phi_by_unit_cell_parameters(
        unit_cell_parameters, flag_unit_cell_parameters=flag_unit_cell_parameters) 

    zeros = numpy.zeros_like(unit_cell_parameters[0])

    # inversed M matrix
    m_inv_m_ij = numpy.stack([
        s_a/(a*phi), zeros, zeros,
        (c_a*c_b-c_g)/(b*phi*s_a), 1/(b*s_a), zeros,
        (c_a*c_g-c_b)/(c*phi*s_a), -c_a/(c*s_a), 1/c], axis=0)
    dder = {}
    if flag_unit_cell_parameters:
        der_11 = numpy.stack([zeros, zeros, zeros, zeros, zeros, zeros], axis=0)
        der_12 = numpy.stack([zeros, zeros, zeros, zeros, zeros, zeros], axis=0)
        der_13 = numpy.stack([zeros, zeros, zeros, zeros, zeros, zeros], axis=0)
        der_21 = numpy.stack([zeros, zeros, zeros, zeros, zeros, zeros], axis=0)
        der_22 = numpy.stack([zeros, zeros, zeros, zeros, zeros, zeros], axis=0)
        der_23 = numpy.stack([zeros, zeros, zeros, zeros, zeros, zeros], axis=0)
        der_31 = numpy.stack([zeros, zeros, zeros, zeros, zeros, zeros], axis=0)
        der_32 = numpy.stack([zeros, zeros, zeros, zeros, zeros, zeros], axis=0)
        der_33 = numpy.stack([zeros, zeros, zeros, zeros, zeros, zeros], axis=0)

        dder["unit_cell_parameters"] = numpy.stack([
            der_11, der_12, der_13, der_21, der_22, der_23, der_31, der_32,
            der_33], axis=0)
    return m_inv_m_ij, dder


def calc_m_m_norm_by_unit_cell_parameters(
        unit_cell_parameters, flag_unit_cell_parameters: bool = False):
    """nM matrix."""
    a, b = unit_cell_parameters[0], unit_cell_parameters[1]
    c = unit_cell_parameters[2]
    alpha, beta = unit_cell_parameters[3], unit_cell_parameters[4]
    gamma = unit_cell_parameters[5]
    c_a, c_b, c_g = numpy.cos(alpha), numpy.cos(beta), numpy.cos(gamma)
    s_a, s_b, s_g = numpy.sin(alpha), numpy.sin(beta), numpy.sin(gamma)

    phi, dder_phi = calc_phi_by_unit_cell_parameters(
        unit_cell_parameters, flag_unit_cell_parameters=flag_unit_cell_parameters) 

    zeros = numpy.zeros_like(unit_cell_parameters[0])
    ones = numpy.ones_like(unit_cell_parameters[0])

    m_m_norm_ij = numpy.stack([
        phi/s_a, zeros, zeros,
        (c_g-c_a*c_b)/s_a, s_a, zeros,
        c_b, c_a, ones], axis=0)

    dder = {}
    if flag_unit_cell_parameters:
        s_a_sq = numpy.square(s_a)
        c_a_sq = numpy.square(c_a)
        der_11 = numpy.stack([
            zeros, zeros, zeros,
            -0.5/phi * (2*c_a-2.*c_b*c_g) - phi*c_a/s_a_sq,
            -0.5/phi * (2*c_b-2.*c_a*c_g)*s_b/s_a,
            -0.5/phi * (2*c_g-2.*c_a*c_b)*s_g/s_a], axis=0)
        der_12 = numpy.stack([zeros, zeros, zeros, zeros, zeros, zeros], axis=0)
        der_13 = numpy.stack([zeros, zeros, zeros, zeros, zeros, zeros], axis=0)
        der_21 = numpy.stack([
            zeros, zeros, zeros, c_b-(c_g*c_a-c_a_sq*c_b)/s_a_sq, c_a*s_b/s_a,
            -s_g/s_a], axis=0)
        der_22 = numpy.stack([zeros, zeros, zeros, c_a, zeros, zeros], axis=0)
        der_23 = numpy.stack([zeros, zeros, zeros, zeros, zeros, zeros], axis=0)
        der_31 = numpy.stack([zeros, zeros, zeros, zeros, -1*s_b, zeros], axis=0)
        der_32 = numpy.stack([zeros, zeros, zeros, -1*s_a, zeros, zeros], axis=0)
        der_33 = numpy.stack([zeros, zeros, zeros, zeros, zeros, zeros], axis=0)

        dder["unit_cell_parameters"] = numpy.stack([
            der_11, der_12, der_13, der_21, der_22, der_23, der_31, der_32,
            der_33], axis=0)
    return m_m_norm_ij, dder


def calc_m_inv_m_norm_by_unit_cell_parameters(
        unit_cell_parameters, flag_unit_cell_parameters: bool = False):
    """nM matrix."""
    a, b = unit_cell_parameters[0], unit_cell_parameters[1]
    c = unit_cell_parameters[2]
    alpha, beta = unit_cell_parameters[3], unit_cell_parameters[4]
    gamma = unit_cell_parameters[5]
    c_a, c_b, c_g = numpy.cos(alpha), numpy.cos(beta), numpy.cos(gamma)
    s_a, s_b, s_g = numpy.sin(alpha), numpy.sin(beta), numpy.sin(gamma)

    phi, dder_phi = calc_phi_by_unit_cell_parameters(
        unit_cell_parameters, flag_unit_cell_parameters=flag_unit_cell_parameters) 

    zeros = numpy.zeros_like(unit_cell_parameters[0])
    ones = numpy.ones_like(unit_cell_parameters[0])

    m_inv_m_norm_ij = numpy.stack([
        s_a/phi, zeros, zeros,
        (c_a*c_b-c_g)/(phi*s_a), 1./s_a, zeros,
        (c_a*c_g-c_b)/(phi*s_a), -c_a/s_a, ones], axis=0)

    dder = {}
    if flag_unit_cell_parameters:
        der_11 = numpy.stack([zeros, zeros, zeros, zeros, zeros, zeros], axis=0)
        der_12 = numpy.stack([zeros, zeros, zeros, zeros, zeros, zeros], axis=0)
        der_13 = numpy.stack([zeros, zeros, zeros, zeros, zeros, zeros], axis=0)
        der_21 = numpy.stack([zeros, zeros, zeros, zeros, zeros, zeros], axis=0)
        der_22 = numpy.stack([zeros, zeros, zeros, zeros, zeros, zeros], axis=0)
        der_23 = numpy.stack([zeros, zeros, zeros, zeros, zeros, zeros], axis=0)
        der_31 = numpy.stack([zeros, zeros, zeros, zeros, zeros, zeros], axis=0)
        der_32 = numpy.stack([zeros, zeros, zeros, zeros, zeros, zeros], axis=0)
        der_33 = numpy.stack([zeros, zeros, zeros, zeros, zeros, zeros], axis=0)

        dder["unit_cell_parameters"] = numpy.stack([
            der_11, der_12, der_13, der_21, der_22, der_23, der_31, der_32,
            der_33], axis=0)
    return m_inv_m_norm_ij, dder


def calc_q_ccs_by_unit_cell_parameters(
        index_hkl, unit_cell_parameters,
        flag_unit_cell_parameters: bool = False):
    """
    Calculate scattering vector in Cartesian coordinate system.
    Coordinate system is defined as (x||a*, z||c).

    Keyword arguments:

        index_hkl: Miller indices (3, refln)

    Output arguments:

        k_x, k_y, k_z: 1D numpy array of x, y, z components of unity
                       scattering vector
    """
    m_b, dder_m_b = calc_m_b_by_unit_cell_parameters(
        unit_cell_parameters, flag_unit_cell_parameters = flag_unit_cell_parameters)
    q_xyz = numpy.stack(
        [m_b[0]*index_hkl[0]+m_b[1]*index_hkl[1]+m_b[2]*index_hkl[2],
         m_b[3]*index_hkl[0]+m_b[4]*index_hkl[1]+m_b[5]*index_hkl[2],
         m_b[6]*index_hkl[0]+m_b[7]*index_hkl[1]+m_b[8]*index_hkl[2]],
        axis=0)

    dder = {}
    if flag_unit_cell_parameters:
        dder["unit_cell_parameters"] = numpy.stack([
            numpy.expand_dims(index_hkl[0], axis=-1)*numpy.expand_dims(dder_m_b["unit_cell_parameters"][0], axis=0) + 
            numpy.expand_dims(index_hkl[1], axis=-1)*numpy.expand_dims(dder_m_b["unit_cell_parameters"][1], axis=0) +
            numpy.expand_dims(index_hkl[2], axis=-1)*numpy.expand_dims(dder_m_b["unit_cell_parameters"][2], axis=0),
            numpy.expand_dims(index_hkl[0], axis=-1)*numpy.expand_dims(dder_m_b["unit_cell_parameters"][3], axis=0) + 
            numpy.expand_dims(index_hkl[1], axis=-1)*numpy.expand_dims(dder_m_b["unit_cell_parameters"][4], axis=0) +
            numpy.expand_dims(index_hkl[2], axis=-1)*numpy.expand_dims(dder_m_b["unit_cell_parameters"][5], axis=0),
            numpy.expand_dims(index_hkl[0], axis=-1)*numpy.expand_dims(dder_m_b["unit_cell_parameters"][6], axis=0) + 
            numpy.expand_dims(index_hkl[1], axis=-1)*numpy.expand_dims(dder_m_b["unit_cell_parameters"][7], axis=0) +
            numpy.expand_dims(index_hkl[2], axis=-1)*numpy.expand_dims(dder_m_b["unit_cell_parameters"][8], axis=0)
        ], axis=0)
    return q_xyz, dder


def calc_eq_ccs_by_unit_cell_parameters(
        index_hkl, unit_cell_parameters,
        flag_unit_cell_parameters: bool = False):
    """
    Calculate unity scattering vector in Cartesian coordinate system.
    Coordinate system is defined as (x||a*, z||c).

    Keyword arguments:

        index_hkl: Miller indices (3, refln)

    Output arguments:

        k_x, k_y, k_z: 1D numpy array of x, y, z components of unity
                       scattering vector
    """
    q_xyz, dder_qxyz = calc_q_ccs_by_unit_cell_parameters(
        index_hkl, unit_cell_parameters, flag_unit_cell_parameters=flag_unit_cell_parameters)

    q_norm = numpy.sqrt(numpy.sum(numpy.square(q_xyz), axis=0))

    q_norm = numpy.where(q_norm != 0., q_norm, 1.)

    eq_xyz = q_xyz/q_norm
    dder = {}
    if flag_unit_cell_parameters:
        q_norm_exp = numpy.expand_dims(numpy.expand_dims(q_norm, axis=0), axis=-1)
        q_xyz_exp = numpy.expand_dims(q_xyz, axis=-1)
        hh = (q_xyz_exp*dder_qxyz["unit_cell_parameters"]).sum(axis=0)
        hh_exp = numpy.expand_dims(hh, axis=0)
        dder["unit_cell_parameters"] = (
            dder_qxyz["unit_cell_parameters"]/q_norm_exp-
            q_xyz_exp/numpy.square(q_norm_exp)*hh_exp/q_norm_exp)
    return eq_xyz, dder


def calc_inv_d_by_unit_cell_parameters(
        index_hkl, unit_cell_parameters,
        flag_unit_cell_parameters: bool = False):
    """
    Calculate 1/d for given reflections hkl
    and unit cell parameters defined as [a, b, c, alpha, beta, gamma].
    """
    a, b = unit_cell_parameters[0], unit_cell_parameters[1]
    c = unit_cell_parameters[2]
    c_a = numpy.cos(unit_cell_parameters[3])
    c_b = numpy.cos(unit_cell_parameters[4])
    c_g = numpy.cos(unit_cell_parameters[5])
    c_a_sq, c_b_sq = numpy.square(c_a), numpy.square(c_b)
    c_g_sq = numpy.square(c_g)
    s_a_sq, s_b_sq, s_g_sq = (1.-c_a_sq), (1.-c_b_sq), (1.-c_g_sq)

    A = (1.-c_a_sq-c_b_sq-c_g_sq+2.*c_a*c_b*c_g)
    h = index_hkl[0]
    k = index_hkl[1]
    l = index_hkl[2]
    B1 = s_a_sq*numpy.square(h*1./a)+\
         s_b_sq*numpy.square(k*1./b)+\
         s_g_sq*numpy.square(l*1./c)
    B2 = 2.*(k*l*(c_a-c_b*c_g))/(b*c)+\
         2.*(h*l*(c_b-c_a*c_g))/(a*c)+\
         2.*(h*k*(c_g-c_a*c_b))/(a*b)
    B = B1-B2
    inv_d = numpy.sqrt(B*1./A)
    dder = {}
    if flag_unit_cell_parameters:
        coeff = 1./(inv_d * A)
        s_a = numpy.sin(unit_cell_parameters[3])
        s_b = numpy.sin(unit_cell_parameters[4])
        s_g = numpy.sin(unit_cell_parameters[5])
        dder_a = coeff * (h/numpy.square(a)) * ((c_b*l)/c + (c_g*k)/b - (s_a_sq*h)/a)*numpy.ones_like(a)
        dder_b = coeff * (k/numpy.square(b)) * ((c_g*h)/a + (c_a*l)/c - (s_b_sq*k)/b)*numpy.ones_like(b)
        dder_c = coeff * (l/numpy.square(c)) * ((c_a*k)/b + (c_b*h)/a - (s_g_sq*l)/c)*numpy.ones_like(c)
        dder_al = (s_a/A) * ((c_a*numpy.square(h/a) + (k*l)/(b*c))/inv_d - inv_d*(c_a - c_b * c_g)/A)*numpy.ones_like(c_a)
        dder_be = (s_b/A) * ((c_b*numpy.square(k/b) + (l*h)/(c*a))/inv_d - inv_d*(c_b - c_g * c_a)/A)*numpy.ones_like(c_b)
        dder_ga = (s_g/A) * ((c_g*numpy.square(l/c) + (h*k)/(a*b))/inv_d - inv_d*(c_g - c_a * c_b)/A)*numpy.ones_like(c_g)
        dder["unit_cell_parameters"] = numpy.stack([dder_a, dder_b, dder_c, dder_al, dder_be, dder_ga], axis=0)
    return inv_d, dder


def calc_sthovl_by_unit_cell_parameters(index_hkl, unit_cell_parameters,
        flag_unit_cell_parameters: bool = False):
    """
    Calculate sin(theta)/lambda for given reflections h, k, l
    and unit cell parameters defined as a, b, c, cos(alpha), cos(beta), cos(gamma).
    """
    inv_d, dder_inv_d = calc_inv_d_by_unit_cell_parameters(
        index_hkl, unit_cell_parameters,
        flag_unit_cell_parameters=flag_unit_cell_parameters)
    sthovl = 0.5*inv_d
    dder = {}
    if flag_unit_cell_parameters:
        dder["unit_cell_parameters"] = 0.5*dder_inv_d["unit_cell_parameters"]
    return sthovl, dder



def transform_quadratic_form_reciprocal_to_ccs(
    quadratic_form_reciprocal, unit_cell_parameters,
    flag_quadratic_form_reciprocal:bool=False,
    flag_unit_cell_parameters:bool=False):
    """Transform quadratic form from reciprocal system
    ($a^*$, $b^*$, $c^*$) to Cartesian coordinate system (x||a*, z||c)
    """
    m_m_ij, dder_minvb = calc_m_inv_b_by_unit_cell_parameters(
        unit_cell_parameters, flag_unit_cell_parameters=flag_unit_cell_parameters)
    m_mt_ij = numpy.stack([
        m_m_ij[0], m_m_ij[3], m_m_ij[6],
        m_m_ij[1], m_m_ij[4], m_m_ij[7],
        m_m_ij[2], m_m_ij[5], m_m_ij[8]], axis=0)
    quadratic_form_ccs, dder_q = calc_m_q_mt(m_mt_ij, quadratic_form_reciprocal,
        flag_m=flag_unit_cell_parameters, flag_q=flag_quadratic_form_reciprocal)
    dder = {}
    if flag_quadratic_form_reciprocal:
        dder["quadratic_form_reciprocal"] = numpy.zeros((6, )+quadratic_form_reciprocal.shape, dtype=quadratic_form_reciprocal.dtype)
    if flag_unit_cell_parameters:
        dder["unit_cell_parameters"] = numpy.zeros((6, )+unit_cell_parameters.shape, dtype=unit_cell_parameters.dtype)
    return quadratic_form_ccs, dder


def transform_quadratic_form_ccs_to_reciprocal(
    quadratic_form_ccs, unit_cell_parameters,
    flag_quadratic_form_ccs:bool=False,
    flag_unit_cell_parameters:bool=False):
    """Transform quadratic form from Cartesian coordinate system (x||a*, z||c)
   to reciprocal system ($a^*$, $b^*$, $c^*$)
    """
    m_m_ij, dder = calc_m_b_by_unit_cell_parameters(
        unit_cell_parameters, flag_unit_cell_parameters=False)
    m_mt_ij = numpy.stack([
        m_m_ij[0], m_m_ij[3], m_m_ij[6],
        m_m_ij[1], m_m_ij[4], m_m_ij[7],
        m_m_ij[2], m_m_ij[5], m_m_ij[8]], axis=0)
    quadratic_form_reciprocal = calc_m_q_mt(m_mt_ij, quadratic_form_ccs, flag_m=False, flag_q=False)[0]
    dder = {}
    if flag_quadratic_form_ccs:
        dder["quadratic_form_ccs"] = numpy.zeros((6, )+quadratic_form_ccs.shape, dtype=quadratic_form_ccs.dtype)
    if flag_unit_cell_parameters:
        dder["unit_cell_parameters"] = numpy.zeros((6, )+unit_cell_parameters.shape, dtype=unit_cell_parameters.dtype)
    return quadratic_form_reciprocal, dder


def transform_quadratic_form_reciprocal_norm_to_ccs(
    quadratic_form_reciprocal_norm, unit_cell_parameters,
    flag_quadratic_form_reciprocal_norm:bool=False,
    flag_unit_cell_parameters:bool=False):
    """Transform quadratic form from normalized reciprocal system
    ($a^*/a^*$, $b^*/b^*$, $c^*/c^*$) to Cartesian coordinate system (x||a*, z||c)
    """
    m_m_ij, dder = calc_m_inv_b_norm_by_unit_cell_parameters(
        unit_cell_parameters, flag_unit_cell_parameters=False)
    m_mt_ij = numpy.stack([
        m_m_ij[0], m_m_ij[3], m_m_ij[6],
        m_m_ij[1], m_m_ij[4], m_m_ij[7],
        m_m_ij[2], m_m_ij[5], m_m_ij[8]], axis=0)
    quadratic_form_ccs = calc_m_q_mt(m_mt_ij, quadratic_form_reciprocal_norm, flag_m=False, flag_q=False)[0]
    dder = {}
    if flag_quadratic_form_reciprocal_norm:
        dder["quadratic_form_reciprocal_norm"] = numpy.zeros((6, )+quadratic_form_reciprocal_norm.shape, dtype=quadratic_form_reciprocal_norm.dtype)
    if flag_unit_cell_parameters:
        dder["unit_cell_parameters"] = numpy.zeros((6, )+unit_cell_parameters.shape, dtype=unit_cell_parameters.dtype)
    return quadratic_form_ccs, dder


def transform_quadratic_form_ccs_to_reciprocal_norm(
    quadratic_form_ccs, unit_cell_parameters,
    flag_quadratic_form_ccs:bool=False,
    flag_unit_cell_parameters:bool=False):
    """Transform quadratic form from Cartesian coordinate system (x||a*, z||c)
   to normalized reciprocal system ($a^*/a^*$, $b^*/b^*$, $c^*/c^*$)
    """
    m_m_ij, dder = calc_m_b_norm_by_unit_cell_parameters(
        unit_cell_parameters, flag_unit_cell_parameters=False)
    m_mt_ij = numpy.stack([
        m_m_ij[0], m_m_ij[3], m_m_ij[6],
        m_m_ij[1], m_m_ij[4], m_m_ij[7],
        m_m_ij[2], m_m_ij[5], m_m_ij[8]], axis=0)
    quadratic_form_reciprocal_norm = calc_m_q_mt(m_mt_ij, quadratic_form_ccs, flag_m=False, flag_q=False)[0]
    dder = {}
    if flag_quadratic_form_ccs:
        dder["quadratic_form_ccs"] = numpy.zeros((6, )+quadratic_form_ccs.shape, dtype=quadratic_form_ccs.dtype)
    if flag_unit_cell_parameters:
        dder["unit_cell_parameters"] = numpy.zeros((6, )+unit_cell_parameters.shape, dtype=unit_cell_parameters.dtype)
    return quadratic_form_reciprocal_norm, dder


def transform_quadratic_form_direct_to_ccs(
    quadratic_form_direct, unit_cell_parameters,
    flag_quadratic_form_direct:bool=False,
    flag_unit_cell_parameters:bool=False):
    """Transform quadratic form from direct system
    ($a$, $b$, $c$) to Cartesian coordinate system (x||a*, z||c)
    """
    m_m_ij, dder = calc_m_inv_m_by_unit_cell_parameters(
        unit_cell_parameters, flag_unit_cell_parameters=False)
    m_mt_ij = numpy.stack([
        m_m_ij[0], m_m_ij[3], m_m_ij[6],
        m_m_ij[1], m_m_ij[4], m_m_ij[7],
        m_m_ij[2], m_m_ij[5], m_m_ij[8]], axis=0)
    quadratic_form_ccs = calc_m_q_mt(m_mt_ij, quadratic_form_direct, flag_m=False, flag_q=False)[0]
    dder = {}
    if flag_quadratic_form_direct:
        dder["quadratic_form_direct"] = numpy.zeros((6, )+quadratic_form_direct.shape, dtype=quadratic_form_direct.dtype)
    if flag_unit_cell_parameters:
        dder["unit_cell_parameters"] = numpy.zeros((6, )+unit_cell_parameters.shape, dtype=unit_cell_parameters.dtype)
    return quadratic_form_ccs, dder


def transform_quadratic_form_ccs_to_direct(
    quadratic_form_ccs, unit_cell_parameters,
    flag_quadratic_form_ccs:bool=False,
    flag_unit_cell_parameters:bool=False):
    """Transform quadratic form from Cartesian coordinate system (x||a*, z||c)
   to direct system ($a$, $b$, $c$)
    """
    m_m_ij, dder = calc_m_m_by_unit_cell_parameters(
        unit_cell_parameters, flag_unit_cell_parameters=False)
    m_mt_ij = numpy.stack([
        m_m_ij[0], m_m_ij[3], m_m_ij[6],
        m_m_ij[1], m_m_ij[4], m_m_ij[7],
        m_m_ij[2], m_m_ij[5], m_m_ij[8]], axis=0)
    quadratic_form_direct = calc_m_q_mt(m_mt_ij, quadratic_form_ccs, flag_m=False, flag_q=False)[0]
    dder = {}
    if flag_quadratic_form_ccs:
        dder["quadratic_form_ccs"] = numpy.zeros((6, )+quadratic_form_ccs.shape, dtype=quadratic_form_ccs.dtype)
    if flag_unit_cell_parameters:
        dder["unit_cell_parameters"] = numpy.zeros((6, )+unit_cell_parameters.shape, dtype=unit_cell_parameters.dtype)
    return quadratic_form_direct, dder


def transform_quadratic_form_direct_norm_to_ccs(
    quadratic_form_direct_norm, unit_cell_parameters,
    flag_quadratic_form_direct_norm:bool=False,
    flag_unit_cell_parameters:bool=False):
    """Transform quadratic form from normalized direct system
    ($a/a$, $b/b$, $c/c$) to Cartesian coordinate system (x||a*, z||c)
    """
    m_m_ij, dder = calc_m_inv_m_norm_by_unit_cell_parameters(
        unit_cell_parameters, flag_unit_cell_parameters=False)
    m_mt_ij = numpy.stack([
        m_m_ij[0], m_m_ij[3], m_m_ij[6],
        m_m_ij[1], m_m_ij[4], m_m_ij[7],
        m_m_ij[2], m_m_ij[5], m_m_ij[8]], axis=0)
    quadratic_form_ccs = calc_m_q_mt(m_mt_ij, quadratic_form_direct_norm, flag_m=False, flag_q=False)[0]
    dder = {}
    if flag_quadratic_form_direct_norm:
        dder["quadratic_form_direct_norm"] = numpy.zeros((6, )+quadratic_form_direct_norm.shape, dtype=quadratic_form_direct_norm.dtype)
    if flag_unit_cell_parameters:
        dder["unit_cell_parameters"] = numpy.zeros((6, )+unit_cell_parameters.shape, dtype=unit_cell_parameters.dtype)
    return quadratic_form_ccs, dder


def transform_quadratic_form_ccs_to_direct_norm(
    quadratic_form_ccs, unit_cell_parameters,
    flag_quadratic_form_ccs:bool=False,
    flag_unit_cell_parameters:bool=False):
    """Transform quadratic form from Cartesian coordinate system (x||a*, z||c)
   to normalized direct system ($a/a$, $b/b$, $c/c$)
    """
    m_m_ij, dder = calc_m_m_norm_by_unit_cell_parameters(
        unit_cell_parameters, flag_unit_cell_parameters=False)
    m_mt_ij = numpy.stack([
        m_m_ij[0], m_m_ij[3], m_m_ij[6],
        m_m_ij[1], m_m_ij[4], m_m_ij[7],
        m_m_ij[2], m_m_ij[5], m_m_ij[8]], axis=0)
    quadratic_form_direct_norm = calc_m_q_mt(m_mt_ij, quadratic_form_ccs, flag_m=False, flag_q=False)[0]
    dder = {}
    if flag_quadratic_form_ccs:
        dder["quadratic_form_ccs"] = numpy.zeros((6, )+quadratic_form_ccs.shape, dtype=quadratic_form_ccs.dtype)
    if flag_unit_cell_parameters:
        dder["unit_cell_parameters"] = numpy.zeros((6, )+unit_cell_parameters.shape, dtype=unit_cell_parameters.dtype)
    return quadratic_form_direct_norm, dder

def calc_m_inv_m_b_norm_by_unit_cell_parameters(
        unit_cell_parameters, flag_unit_cell_parameters: bool = False):
    """
    Calculation M^-1 B_norm
    """
    a, b, c = unit_cell_parameters[0], unit_cell_parameters[1], unit_cell_parameters[2]
    al, be, ga = unit_cell_parameters[3], unit_cell_parameters[4], unit_cell_parameters[5]
    c_a, c_b, c_g = numpy.cos(al), numpy.cos(be), numpy.cos(ga)
    s_a, s_b, s_g = numpy.sin(al), numpy.sin(be), numpy.sin(ga)
    phi, dder_phi = calc_phi_by_unit_cell_parameters(
        unit_cell_parameters,
        flag_unit_cell_parameters=flag_unit_cell_parameters)

    o_11 = s_a / (a * phi)
    o_12 = (c_a * c_b - c_g) / (a * phi * s_b)
    o_13 = (c_a * c_g - c_b) / (a * phi * s_g)
    o_21 = (c_a * c_b - c_g) / (b * phi * s_a)
    o_22 = s_b / (b * phi)
    o_23 = (c_b * c_g - c_a) / (b * phi * s_g)
    o_31 = (c_a * c_g - c_b) / (c * phi * s_a)
    o_32 = (c_b * c_g - c_a) / (c * phi * s_b)
    o_33 = s_g / (c * phi)

    inv_m_b_norm = numpy.stack([
        o_11, o_12, o_13, o_21, o_22, o_23, o_31, o_32, o_33],
        axis=0)
    dder = {}
    if flag_unit_cell_parameters:
        zeros = numpy.zeros_like(a)
        der_11 = numpy.stack([zeros, zeros, zeros, zeros, zeros, zeros], axis=0)
        der_12 = numpy.stack([zeros, zeros, zeros, zeros, zeros, zeros], axis=0)
        der_13 = numpy.stack([zeros, zeros, zeros, zeros, zeros, zeros], axis=0)
        der_21 = numpy.stack([zeros, zeros, zeros, zeros, zeros, zeros], axis=0)
        der_22 = numpy.stack([zeros, zeros, zeros, zeros, zeros, zeros], axis=0)
        der_23 = numpy.stack([zeros, zeros, zeros, zeros, zeros, zeros], axis=0)
        der_31 = numpy.stack([zeros, zeros, zeros, zeros, zeros, zeros], axis=0)
        der_32 = numpy.stack([zeros, zeros, zeros, zeros, zeros, zeros], axis=0)
        der_33 = numpy.stack([zeros, zeros, zeros, zeros, zeros, zeros], axis=0)
        dder["unit_cell_parameters"] = numpy.stack([
            der_11, der_12, der_13, der_21, der_22, der_23, der_31, der_32, der_33], axis=0)
    return inv_m_b_norm, dder

def calc_m_inv_b_norm_m_by_unit_cell_parameters(
        unit_cell_parameters, flag_unit_cell_parameters: bool = False):
    """
    Calculation B^-1_norm M
    """
    a, b, c = unit_cell_parameters[0], unit_cell_parameters[1], unit_cell_parameters[2]
    al, be, ga = unit_cell_parameters[3], unit_cell_parameters[4], unit_cell_parameters[5]
    c_a, c_b, c_g = numpy.cos(al), numpy.cos(be), numpy.cos(ga)
    s_a, s_b, s_g = numpy.sin(al), numpy.sin(be), numpy.sin(ga)
    phi, dder_phi = calc_phi_by_unit_cell_parameters(
        unit_cell_parameters,
        flag_unit_cell_parameters=flag_unit_cell_parameters)
    o_11 = a * s_a / phi
    o_12 = b * s_a * c_g / phi
    o_13 = c * s_a * c_b / phi
    o_21 = a * s_b * c_g / phi
    o_22 = b * s_b / phi
    o_23 = c * s_b * c_a / phi
    o_31 = a * s_g * c_b / phi
    o_32 = b * s_g * c_a / phi
    o_33 = c * s_g / phi

    inv_b_norm_m = numpy.stack([
        o_11, o_12, o_13, o_21, o_22, o_23, o_31, o_32, o_33],
        axis=0)
    dder = {}
    if flag_unit_cell_parameters:
        zeros = numpy.zeros_like(a)
        der_11 = numpy.stack([zeros, zeros, zeros, zeros, zeros, zeros], axis=0)
        der_12 = numpy.stack([zeros, zeros, zeros, zeros, zeros, zeros], axis=0)
        der_13 = numpy.stack([zeros, zeros, zeros, zeros, zeros, zeros], axis=0)
        der_21 = numpy.stack([zeros, zeros, zeros, zeros, zeros, zeros], axis=0)
        der_22 = numpy.stack([zeros, zeros, zeros, zeros, zeros, zeros], axis=0)
        der_23 = numpy.stack([zeros, zeros, zeros, zeros, zeros, zeros], axis=0)
        der_31 = numpy.stack([zeros, zeros, zeros, zeros, zeros, zeros], axis=0)
        der_32 = numpy.stack([zeros, zeros, zeros, zeros, zeros, zeros], axis=0)
        der_33 = numpy.stack([zeros, zeros, zeros, zeros, zeros, zeros], axis=0)
        dder["unit_cell_parameters"] = numpy.stack([
            der_11, der_12, der_13, der_21, der_22, der_23, der_31, der_32, der_33], axis=0)
    return inv_b_norm_m, dder


def calc_matrix_t(index_hkl, unit_cell_parameters, flag_unit_cell_parameters: bool = False):
    """Calculate matrix T to have new z axis along scattering vector q = h a* + k b* + l c* from CCS in which (x||a*, z||c).

    m is vector in Cartesian coordinate system (x||a*, z||c)
    M is vector in Cartesian coordinate system (Z||q, axes X and Y are not important)

    M = T * m
    """
    eq_ccs, dder_eq_ccs = calc_eq_ccs_by_unit_cell_parameters(
        index_hkl, unit_cell_parameters, flag_unit_cell_parameters=flag_unit_cell_parameters)

    k_x = eq_ccs[0]
    k_y = eq_ccs[1]
    k_z = eq_ccs[2]
    
    al = numpy.zeros_like(k_x, dtype=float)
    be = numpy.arccos(k_z)
    sb = numpy.sin(be)
    flag = (sb != 0.)
    sa1 = k_x[flag]*1./sb[flag]
    ca2 = -1*k_y[flag]*1./sb[flag]
    sa1[sa1 > 1] = 1.
    sa1[sa1 < -1] = -1.
    ca2[ca2 > 1] = 1.
    ca2[ca2 < -1] = -1.
    al1 = numpy.arcsin(sa1)
    al2 = numpy.arccos(ca2)
    al_sh = numpy.copy(al1)
    al_sh[sa1 > 0.] = al2[sa1 > 0.]
    al_sh[sa1 <= 0.] = 2.*numpy.pi-al2[sa1 <= 0.]
    al_sh[numpy.abs(al2-al1) < 0.00001] = al1[numpy.abs(al2-al1) < 0.00001]
    al[flag] = al_sh
    ga = 0.
    ca, cb, cg = numpy.cos(al), numpy.cos(be), numpy.cos(ga)
    sa, sb, sg = numpy.sin(al), numpy.sin(be), numpy.sin(ga)

    # here is taken transposed one (== inversed one) to get relation M_(Z||hkl) = T * m_(x||a*, z||c)
    t_11, t_21, t_31 = ca*cg-sa*cb*sg, -ca*sg-sa*cb*cg,  sa*sb
    t_12, t_22, t_32 = sa*cg+ca*cb*sg, -sa*sg+ca*cb*cg, -ca*sb
    t_13, t_23, t_33 = sb*sg, sb*cg, cb
    flag = (((sa*sb-k_x)**2+(-ca*sb-k_y)**2+(cb-k_z)**2) > 0.0001)
    if any(flag):
        warnings.warn("Mistake with eq\nProgram is stopped",
                      UserWarning, stacklevel=2)
        quit()
    m_t_ij  = numpy.stack([t_11, t_12, t_13, t_21, t_22, t_23, t_31, t_32, t_33], axis=0)

    dder = {}
    if flag_unit_cell_parameters:
        dder["unit_cell_parameters"] = numpy.zeros((9, ) + unit_cell_parameters.shape, dtype=unit_cell_parameters.dtype)
    return m_t_ij, dder


from typing import NoReturn
import numpy
from fractions import Fraction

from cryspy.A_functions_base.function_1_objects import \
    form_items_by_dictionary
from cryspy.A_functions_base.function_1_strings import \
    transform_string_to_r_b, transform_r_b_to_string
from cryspy.A_functions_base.function_2_space_group import \
    get_shift_by_centring_type, is_good_for_mask

from cryspy.B_parent_classes.cl_1_item import ItemN
from cryspy.B_parent_classes.cl_2_loop import LoopN


class SpaceGroupWyckoff(ItemN):
    """Wyckoff position.
    
    Contains information about Wyckoff positions of a space group.
    Only one site can be given for each special position but the
    remainder can be generated by applying the symmetry operations
    stored in _space_group_symop.operation_xyz.

    """
    ATTR_MANDATORY_NAMES = ("coord_xyz", )
    ATTR_MANDATORY_TYPES = (str, )
    ATTR_MANDATORY_CIF = ("coord_xyz", )

    ATTR_OPTIONAL_NAMES = ("id", "letter", "multiplicity", "site_symmetry")
    ATTR_OPTIONAL_TYPES = (str, str, int, str)
    ATTR_OPTIONAL_CIF = ("id", "letter", "multiplicity", "site_symmetry")

    ATTR_NAMES = ATTR_MANDATORY_NAMES + ATTR_OPTIONAL_NAMES
    ATTR_TYPES = ATTR_MANDATORY_TYPES + ATTR_OPTIONAL_TYPES
    ATTR_CIF = ATTR_MANDATORY_CIF + ATTR_OPTIONAL_CIF

    ATTR_INT_NAMES = ("full_coord_xyz", "r", "b", "full_r", "full_b", 
                      "centring_type")
    ATTR_INT_PROTECTED_NAMES = ("it_coord_xyz", )

    # parameters considered are refined parameters
    ATTR_REF = ()
    ATTR_SIGMA = tuple([f"{_h:}_sigma" for _h in ATTR_REF])
    ATTR_CONSTR_FLAG = tuple([f"{_h:}_constraint" for _h in ATTR_REF])
    ATTR_REF_FLAG = tuple([f"{_h:}_refinement" for _h in ATTR_REF])
    ATTR_CONSTR_MARK = tuple([f"{_h:}_mark" for _h in ATTR_REF])

    # constraints on the parameters
    D_CONSTRAINTS = {}

    # default values for the parameters
    D_DEFAULT = {}
    for key in ATTR_SIGMA:
        D_DEFAULT[key] = 0.
    for key in (ATTR_CONSTR_FLAG + ATTR_REF_FLAG):
        D_DEFAULT[key] = False
    for key in ATTR_CONSTR_MARK:
        D_DEFAULT[key] = ""

    PREFIX = "space_group_Wyckoff"

    def __init__(self, **kwargs) -> NoReturn:
        super(SpaceGroupWyckoff, self).__init__()

        # defined for any integer and float parameters
        D_MIN = {}

        # defined for ani integer and float parameters
        D_MAX = {}

        self.__dict__["D_MIN"] = D_MIN
        self.__dict__["D_MAX"] = D_MAX
        for key, attr in self.D_DEFAULT.items():
            setattr(self, key, attr)
        for key, attr in kwargs.items():
            setattr(self, key, attr)

    def form_object(self) -> NoReturn:
        coord_xyz = self.coord_xyz
        if coord_xyz is None:
            return False
        r, b = transform_string_to_r_b(coord_xyz, labels=("x", "y", "z"))
        self.__dict__["r"] = r
        self.__dict__["b"] = b

        it_coord_xyz = self.it_coord_xyz
        centring_type = self.centring_type
        shift = get_shift_by_centring_type(centring_type)

        if it_coord_xyz is not None:
            full_r, full_b = [], []
            full_coord_xyz = []
            for _coord_xyz in it_coord_xyz:
                r, b = transform_string_to_r_b(_coord_xyz, labels=("x", "y", "z"))

                for _shift in shift:
                    b_new = numpy.mod(b + numpy.array(_shift, dtype=Fraction), 1)
                    _symop = transform_r_b_to_string(r, b_new, labels=("x", "y", "z"))
                    full_coord_xyz.append(_symop)
                    full_r.append(r)
                    full_b.append(b_new)

            self.__dict__["full_r"] = full_r
            self.__dict__["full_b"] = full_b
            self.__dict__["full_coord_xyz"] = full_coord_xyz

    def is_valid_for_fract(self, fract_x: float, fract_y: float,
                           fract_z: float, tol=10 ** -5) -> bool:
        fract_x, fract_y = float(fract_x), float(fract_y)
        fract_z = float(fract_z)
        nval = int(tol ** -1)
        flag_res = False
        for r, b in zip(self.full_r, self.full_b):
            flag_res = is_good_for_mask(
                r, b, Fraction(fract_x).limit_denominator(nval),
                Fraction(fract_y).limit_denominator(nval),
                Fraction(fract_z).limit_denominator(nval))
            if flag_res:
                break
        return flag_res

    def give_default_xyz(self, xyz_0):
        one_pm_1 = numpy.array([ 1, 1, 1], dtype=int)
        one_pm_2 = numpy.array([-1, 1, 1], dtype=int)
        one_pm_3 = numpy.array([ 1,-1, 1], dtype=int)
        one_pm_4 = numpy.array([ 1, 1,-1], dtype=int)
        one_pm_5 = numpy.array([ 1,-1,-1], dtype=int)
        one_pm_6 = numpy.array([-1, 1,-1], dtype=int)
        one_pm_7 = numpy.array([-1,-1, 1], dtype=int)
        one_pm_8 = numpy.array([-1,-1,-1], dtype=int)

        b_float = self.b.astype(float)
        r_float = self.r.astype(float)
        _h = None
        np_r = numpy.array(self.full_r, dtype=Fraction)
        np_b = numpy.array(self.full_b, dtype=Fraction)
        np_b_x = np_b[:, 0]
        np_b_y = np_b[:, 1]
        np_b_z = np_b[:, 2]

        np_r_11, np_r_12, np_r_13 = np_r[:, 0, 0], np_r[:, 0, 1], np_r[:, 0, 2]
        np_r_21, np_r_22, np_r_23 = np_r[:, 1, 0], np_r[:, 1, 1], np_r[:, 1, 2]
        np_r_31, np_r_32, np_r_33 = np_r[:, 2, 0], np_r[:, 2, 1], np_r[:, 2, 2]
        nval = 10**5
        x_0 = Fraction(xyz_0[0]).limit_denominator(nval)
        y_0 = Fraction(xyz_0[1]).limit_denominator(nval)
        z_0 = Fraction(xyz_0[2]).limit_denominator(nval)
        l_ind = [(x_0, y_0, z_0), (x_0, z_0, y_0), (y_0, x_0, z_0), (y_0, z_0, x_0), (z_0, x_0, y_0), (z_0, y_0, x_0)]
        for _ind in l_ind:
            x_fr, y_fr, z_fr = _ind[0], _ind[1], _ind[2]
            r_31_x, r_12_y, r_13_z = np_r_31*x_fr, np_r_12*y_fr, np_r_13*z_fr
            r_11_x, r_22_y, r_23_z = np_r_11*x_fr, np_r_22*y_fr, np_r_23*z_fr
            r_21_x, r_32_y, r_33_z = np_r_21*x_fr, np_r_32*y_fr, np_r_33*z_fr

            val_0 = (r_11_x + r_12_y + r_13_z + np_b_x-x_0).astype(float)
            val_1 = (r_21_x + r_22_y + r_23_z + np_b_y-y_0).astype(float)
            val_2 = (r_31_x + r_32_y + r_33_z + np_b_z-z_0).astype(float)
            flag_2d_0 = numpy.isclose(numpy.mod(val_0, 1), 0.)
            flag_2d_1 = numpy.isclose(numpy.mod(val_1, 1), 0.)
            flag_2d_2 = numpy.isclose(numpy.mod(val_2, 1), 0.)
            flag_1 = numpy.any(flag_2d_0*flag_2d_1*flag_2d_2, axis=0)
            if flag_1:
                _h = numpy.array([float(x_fr), float(y_fr), float(z_fr)], dtype=float)
                break

            val_0 = (-r_11_x + r_12_y + r_13_z + np_b_x-x_0).astype(float)
            val_1 = (-r_21_x + r_22_y + r_23_z + np_b_y-y_0).astype(float)
            val_2 = (-r_31_x + r_32_y + r_33_z + np_b_z-z_0).astype(float)
            flag_2d_0 = numpy.isclose(numpy.mod(val_0, 1), 0.)
            flag_2d_1 = numpy.isclose(numpy.mod(val_1, 1), 0.)
            flag_2d_2 = numpy.isclose(numpy.mod(val_2, 1), 0.)
            flag_2 = numpy.any(flag_2d_0*flag_2d_1*flag_2d_2, axis=0)
            if flag_2:
                _h = numpy.array([-float(x_fr), float(y_fr), float(z_fr)], dtype=float)
                break

            val_0 = (r_11_x - r_12_y + r_13_z + np_b_x-x_0).astype(float)
            val_1 = (r_21_x - r_22_y + r_23_z + np_b_y-y_0).astype(float)
            val_2 = (r_31_x - r_32_y + r_33_z + np_b_z-z_0).astype(float)
            flag_2d_0 = numpy.isclose(numpy.mod(val_0, 1), 0.)
            flag_2d_1 = numpy.isclose(numpy.mod(val_1, 1), 0.)
            flag_2d_2 = numpy.isclose(numpy.mod(val_2, 1), 0.)
            flag_3 = numpy.any(flag_2d_0*flag_2d_1*flag_2d_2, axis=0)
            if flag_3:
                _h = numpy.array([float(x_fr), -float(y_fr), float(z_fr)], dtype=float)
                break

            val_0 = (r_11_x + r_12_y - r_13_z + np_b_x-x_0).astype(float)
            val_1 = (r_21_x + r_22_y - r_23_z + np_b_y-y_0).astype(float)
            val_2 = (r_31_x + r_32_y - r_33_z + np_b_z-z_0).astype(float)
            flag_2d_0 = numpy.isclose(numpy.mod(val_0, 1), 0.)
            flag_2d_1 = numpy.isclose(numpy.mod(val_1, 1), 0.)
            flag_2d_2 = numpy.isclose(numpy.mod(val_2, 1), 0.)
            flag_4 = numpy.any(flag_2d_0*flag_2d_1*flag_2d_2, axis=0)
            if flag_4:
                _h = numpy.array([float(x_fr), float(y_fr), -float(z_fr)], dtype=float)
                break

            val_0 = (r_11_x - r_12_y - r_13_z + np_b_x-x_0).astype(float)
            val_1 = (r_21_x - r_22_y - r_23_z + np_b_y-y_0).astype(float)
            val_2 = (r_31_x - r_32_y - r_33_z + np_b_z-z_0).astype(float)
            flag_2d_0 = numpy.isclose(numpy.mod(val_0, 1), 0.)
            flag_2d_1 = numpy.isclose(numpy.mod(val_1, 1), 0.)
            flag_2d_2 = numpy.isclose(numpy.mod(val_2, 1), 0.)
            flag_5 = numpy.any(flag_2d_0*flag_2d_1*flag_2d_2, axis=0)
            if flag_5:
                _h = numpy.array([float(x_fr), -float(y_fr), -float(z_fr)], dtype=float)
                break

            val_0 = (-r_11_x + r_12_y - r_13_z + np_b_x-x_0).astype(float)
            val_1 = (-r_21_x + r_22_y - r_23_z + np_b_y-y_0).astype(float)
            val_2 = (-r_31_x + r_32_y - r_33_z + np_b_z-z_0).astype(float)
            flag_2d_0 = numpy.isclose(numpy.mod(val_0, 1), 0.)
            flag_2d_1 = numpy.isclose(numpy.mod(val_1, 1), 0.)
            flag_2d_2 = numpy.isclose(numpy.mod(val_2, 1), 0.)
            flag_6 = numpy.any(flag_2d_0*flag_2d_1*flag_2d_2, axis=0)
            if flag_6:
                _h = numpy.array([-float(x_fr), float(y_fr), -float(z_fr)], dtype=float)
                break

            val_0 = (-r_11_x - r_12_y + r_13_z + np_b_x-x_0).astype(float)
            val_1 = (-r_21_x - r_22_y + r_23_z + np_b_y-y_0).astype(float)
            val_2 = (-r_31_x - r_32_y + r_33_z + np_b_z-z_0).astype(float)
            flag_2d_0 = numpy.isclose(numpy.mod(val_0, 1), 0.)
            flag_2d_1 = numpy.isclose(numpy.mod(val_1, 1), 0.)
            flag_2d_2 = numpy.isclose(numpy.mod(val_2, 1), 0.)
            flag_7 = numpy.any(flag_2d_0*flag_2d_1*flag_2d_2, axis=0)
            if flag_7:
                _h = numpy.array([-float(x_fr), -float(y_fr), float(z_fr)], dtype=float)
                break

            val_0 = (-r_11_x - r_12_y - r_13_z + np_b_x-x_0).astype(float)
            val_1 = (-r_21_x - r_22_y - r_23_z + np_b_y-y_0).astype(float)
            val_2 = (-r_31_x - r_32_y - r_33_z + np_b_z-z_0).astype(float)
            flag_2d_0 = numpy.isclose(numpy.mod(val_0, 1), 0.)
            flag_2d_1 = numpy.isclose(numpy.mod(val_1, 1), 0.)
            flag_2d_2 = numpy.isclose(numpy.mod(val_2, 1), 0.)
            flag_8 = numpy.any(flag_2d_0*flag_2d_1*flag_2d_2, axis=0)
            if flag_8:
                _h = numpy.array([-float(x_fr), -float(y_fr), -float(z_fr)], dtype=float)
                break
        if _h is None:  # not sure about this condition, but may be it is needed when x,y,z are refined
            _h = numpy.array([float(x_0), float(y_0), float(z_0)], dtype=float)
        xyz_new = (numpy.matmul(r_float, _h) + b_float) % 1
        return xyz_new

class SpaceGroupWyckoffL(LoopN):
    """Wyckoff positions.

    Contains information about Wyckoff positions of a space group.
    Only one site can be given for each special position but the
    remainder can be generated by applying the symmetry operations
    stored in _space_group_symop.operation_xyz.

    """
    ITEM_CLASS = SpaceGroupWyckoff
    ATTR_INDEX = "id"
    def __init__(self, loop_name: str = None, **kwargs) -> NoReturn:
        super(SpaceGroupWyckoffL, self).__init__()
        self.__dict__["items"] = form_items_by_dictionary(self.ITEM_CLASS, kwargs)
        self.__dict__["loop_name"] = loop_name

    def get_id_for_fract(self, fract_x: float, fract_y: float, fract_z: float, tol=10 ** -5) -> str:
        l_res = []
        for _item in self.items:
            if _item.is_valid_for_fract(fract_x, fract_y, fract_z, tol):
                l_res.append((_item.id, _item.multiplicity))
        out = sorted(l_res, key=lambda x: x[1])  # sort by multiplicity
        return out[0][0]

    def get_letter_for_fract(self, fract_x: float, fract_y: float, fract_z: float, tol=10 ** -5) -> str:
        _id = self.get_id_for_fract(fract_x, fract_y, fract_z)
        res = self[_id].letter
        return res

    def get_wyckoff_for_fract(self, fract_x: float, fract_y: float, fract_z: float, tol=10 ** -5) -> str:
        _id = self.get_id_for_fract(fract_x, fract_y, fract_z)
        return self[_id]
    
# s_cont = """
#   loop_
#   _space_group_wyckoff.id
#   _space_group_wyckoff.multiplicity
#   _space_group_wyckoff.letter
#   _space_group_wyckoff.site_symmetry
#   _space_group_wyckoff.coord_xyz
#     1  192   h   1      x,y,z
#     2   96   g   ..2    1/4,y,-y
#     3   96   f   2..    x,1/8,1/8
#     4   32   b   .32    1/4,1/4,1/4
# """

# obj = SpaceGroupWyckoffL.from_cif(s_cont)
# print(obj, end="\n\n")
# print(obj["1"], end="\n\n")



import numpy
from typing import NoReturn

from cryspy.B_parent_classes.cl_1_item import ItemN
from cryspy.B_parent_classes.cl_2_loop import LoopN


class Pd2dInstrResolution(ItemN):
    """
    PdInstrReflexAsymmetry describes asymmetry of Bragg reflections for
    1d powder diffractometer.

    Mandatory attributes:
        - u, v, w, x, y, z

    """
    ATTR_MANDATORY_NAMES = ("u", "v", "w", "x", "y")
    ATTR_MANDATORY_TYPES = (float, float, float, float, float)
    ATTR_MANDATORY_CIF = ("U", "V", "W", "X", "Y")

    ATTR_OPTIONAL_NAMES = ("phi", )
    ATTR_OPTIONAL_TYPES = (float, )
    ATTR_OPTIONAL_CIF = ("phi", )

    ATTR_NAMES = ATTR_MANDATORY_NAMES + ATTR_OPTIONAL_NAMES
    ATTR_TYPES = ATTR_MANDATORY_TYPES + ATTR_OPTIONAL_TYPES
    ATTR_CIF = ATTR_MANDATORY_CIF + ATTR_OPTIONAL_CIF

    ATTR_INT_NAMES = ()
    ATTR_INT_PROTECTED_NAMES = ()

    # parameters considered are refined parameters
    ATTR_REF = ("u", "v", "w", "x", "y", "phi")
    ATTR_SIGMA = tuple([f"{_h:}_sigma" for _h in ATTR_REF])
    ATTR_CONSTR_FLAG = tuple([f"{_h:}_constraint" for _h in ATTR_REF])
    ATTR_REF_FLAG = tuple([f"{_h:}_refinement" for _h in ATTR_REF])
    ATTR_CONSTR_MARK = tuple([f"{_h:}_mark" for _h in ATTR_REF])

    # formats if cif format
    D_FORMATS = {"u": "{:.5f}", "v": "{:.5f}", "w": "{:.5f}", "x": "{:.2f}",
                 "y": "{:.5f}", "phi": "{:.1f}"}

    # constraints on the parameters
    D_CONSTRAINTS = {}

    # default values for the parameters
    D_DEFAULT = {"u": 0., "v": 0., "w": 0., "x": 0., "y": 0., "phi": 0.}
    for key in ATTR_SIGMA:
        D_DEFAULT[key] = 0.
    for key in (ATTR_CONSTR_FLAG + ATTR_REF_FLAG):
        D_DEFAULT[key] = False
    for key in ATTR_CONSTR_MARK:
        D_DEFAULT[key] = ""

    PREFIX = "pd2d_instr_resolution"

    def __init__(self, **kwargs) -> NoReturn:
        super(Pd2dInstrResolution, self).__init__()

        # defined for any integer and float parameters
        D_MIN = {"x": 0., "y": 0.}

        # defined for ani integer and float parameters
        D_MAX = {}

        self.__dict__["D_MIN"] = D_MIN
        self.__dict__["D_MAX"] = D_MAX
        for key, attr in self.D_DEFAULT.items():
            setattr(self, key, attr)
        for key, attr in kwargs.items():
            setattr(self, key, attr)


    def _calc_tancos(self, th_hkl):
        """
        tth_hkl in radianas
        calculate tangenth (theta)
        """
        self.t_th = numpy.tan(th_hkl)
        self.t_th_sq = self.t_th**2
        res = numpy.cos(th_hkl)

        self.c_th = res
        self.ic_th = 1./res

    def _calc_hg(self, phase_igsize: float = 0., phase_u: float = 0.,
                 phase_v: float = 0., phase_w: float = 0.):
        """
        ttheta in radians, could be array
        gauss size
        """
        u, v = float(self.u)+phase_u, float(self.v)+phase_v
        w = float(self.w)+phase_w
        res_sq = (u*self.t_th_sq + v*self.t_th + w + 
                  phase_igsize*self.ic_th**2)
        self.hg = numpy.sqrt(res_sq)

    def _calc_hl(self, phase_x: float = 0., phase_y: float = 0.):
        """
        ttheta in radians, could be array
        lorentz site
        """
        x, y = float(self.x)+phase_x, float(self.y)+phase_y
        self.hl = x*self.t_th + y*self.ic_th

    def _calc_hpveta(self):
        """
        ttheta in radians, could be array
        pseudo-Voight function
        """
        hg = self.hg
        hl = self.hl

        hg_2, hl_2 = hg**2, hl**2
        hg_3, hl_3 = hg_2*hg, hl_2*hl
        hg_4, hl_4 = hg_3*hg, hl_3*hl
        hg_5, hl_5 = hg_4*hg, hl_4*hl
        c_2, c_3, c_4, c_5 = 2.69269, 2.42843, 4.47163, 0.07842
        hpv = (hg_5 + c_2*hg_4*hl + c_3*hg_3*hl_2 + 
                       c_4*hg_2*hl_3 + c_5*hg*hl_4 + hl_5)**0.2
        hh = hl*1./hpv
        self.hpv = hpv 
        self.eta = 1.36603*hh - 0.47719*hh**2 + 0.11116*hh**3

    def _calc_agbg(self):
        hpv = self.hpv

        self.ag = (2./hpv)*(numpy.log(2.)/numpy.pi)**0.5
        self.bg = 4*numpy.log(2)/(hpv**2)
        
    def _calc_albl(self):
        hpv = self.hpv
        self.al = 2./(numpy.pi*hpv )
        self.bl = 4./(hpv**2)
    
    def calc_resolution(
            self, tth_hkl, phase_igsize: float = 0., phase_u: float = 0.,
            phase_v: float = 0., phase_w: float = 0., phase_x: float = 0.,
            phase_y: float = 0.):
        """
Calculate parameters for tth
tth_hkl in degrees
        """
        self._calc_tancos(0.5*tth_hkl*numpy.pi/180.)
        self._calc_hg(phase_igsize=phase_igsize, phase_u=phase_u,
                      phase_v=phase_v, phase_w=phase_w)
        self._calc_hl(phase_x=phase_x, phase_y=phase_y)
        self._calc_hpveta()
        self._calc_agbg()
        self._calc_albl()

        a_g = self.ag  
        b_g = self.bg  
        a_l = self.al  
        b_l = self.bl 
        h_g = self.hg  
        h_l = self.hl 
        h_pv = self.hpv 
        eta = self.eta
        
        return h_pv, eta, h_g, h_l, a_g, b_g, a_l, b_l

    def get_dictionary(self):
        res = {}
        res["resolution_parameters"] = numpy.array([
            self.u, self.v, self.w,
            self.x, self.y], dtype=float)

        res["flags_resolution_parameters"] = numpy.array([
            self.u_refinement, self.v_refinement, self.w_refinement,
            self.x_refinement, self.y_refinement], dtype=bool)

        res["resolution_phi_parameter"] = numpy.array([
            self.phi,], dtype=float)
        res["flags_resolution_phi_parameter"] = numpy.array([
            self.phi_refinement,], dtype=float)
                        
        return res
        
class Pd2dInstrResolutionL(LoopN):
    """
    Description of AtomSite in loop.

    """
    ITEM_CLASS = Pd2dInstrResolution
    ATTR_INDEX = None
    def __init__(self, loop_name = None) -> NoReturn:
        super(Pd2dInstrResolutionL, self).__init__()
        self.__dict__["items"] = []
        self.__dict__["loop_name"] = loop_name
   

# s_cont = """
#   loop_

#   _pd2d_instr_resolution_u 
#   _pd2d_instr_resolution_v 
#   _pd2d_instr_resolution_w
#   _pd2d_instr_resolution_x 
#   _pd2d_instr_resolution_y 
# 16.9776(25) -2.8357 0.5763 0. 0.
# """

# obj = Pd2dInstrResolutionL.from_cif(s_cont)
# print(obj, end="\n\n")



"""DensityPoint and DensityPointL classes."""
from typing import NoReturn
from cryspy.B_parent_classes.cl_1_item import ItemN
from cryspy.B_parent_classes.cl_2_loop import LoopN

from cryspy.A_functions_base.function_1_objects import \
    form_items_by_dictionary

import numpy

class ChannelPlusMinus(ItemN):
    """
    Magnetization density in channel plus minus.

    Attributes
    ----------
        - numerator_x, numerator_y, numerator_z, denominator_xyz (mandatory)
        - index_x, index_y, index_z (optional)
        - density_plus
        - density_minus
        - point_multiplicity (optional)
    """

    ATTR_MANDATORY_NAMES = ("numerator_x", "numerator_y", "numerator_z", "denominator_xyz",
        "density_plus", "density_minus", )
    ATTR_MANDATORY_TYPES = (int, int, int, int,
        float, float, )
    ATTR_MANDATORY_CIF = ("numerator_x", "numerator_y", "numerator_z", "denominator_xyz",
        "density_plus", "density_minus", )

    ATTR_OPTIONAL_NAMES = ("index_x", "index_y", "index_z", "point_multiplicity", )
    ATTR_OPTIONAL_TYPES = (int, int, int, int)
    ATTR_OPTIONAL_CIF = ("index_x", "index_y", "index_z", "point_multiplicity", )

    ATTR_NAMES = ATTR_MANDATORY_NAMES + ATTR_OPTIONAL_NAMES
    ATTR_TYPES = ATTR_MANDATORY_TYPES + ATTR_OPTIONAL_TYPES
    ATTR_CIF = ATTR_MANDATORY_CIF + ATTR_OPTIONAL_CIF

    ATTR_INT_NAMES = ()
    ATTR_INT_PROTECTED_NAMES = ()

    # parameters considered are refined parameters
    ATTR_REF = ()
    ATTR_SIGMA = tuple([f"{_h:}_sigma" for _h in ATTR_REF])
    ATTR_CONSTR_FLAG = tuple([f"{_h:}_constraint" for _h in ATTR_REF])
    ATTR_REF_FLAG = tuple([f"{_h:}_refinement" for _h in ATTR_REF])
    ATTR_CONSTR_MARK = tuple([f"{_h:}_mark" for _h in ATTR_REF])

    # formats if cif format
    D_FORMATS = {"density_plus": "{:.5f}", "density_minus": "{:.5f}"}

    # constraints on the parameters
    D_CONSTRAINTS = {}

    # default values for the parameters
    D_DEFAULT = {}
    for key in ATTR_SIGMA:
        D_DEFAULT[key] = 0.
    for key in (ATTR_CONSTR_FLAG + ATTR_REF_FLAG):
        D_DEFAULT[key] = False
    for key in ATTR_CONSTR_MARK:
        D_DEFAULT[key] = ""

    PREFIX = "channel_plus_minus"

    def __init__(self, **kwargs) -> NoReturn:
        super(ChannelPlusMinus, self).__init__()

        # defined for any integer and float parameters
        D_MIN = {"density_plus": 0., "density_minus": 0.,
                 "point_multiplicity": 0,}

        # defined for ani integer and float parameters
        D_MAX = {}

        self.__dict__["D_MIN"] = D_MIN
        self.__dict__["D_MAX"] = D_MAX
        for key, attr in self.D_DEFAULT.items():
            setattr(self, key, attr)
        for key, attr in kwargs.items():
            setattr(self, key, attr)


class ChannelPlusMinusL(LoopN):
    """
    Magnetization density in channel chi in loop

    Attributes
    ----------
        - numerator_x, numerator_y, numerator_z, denominator_xyz (mandatory)
        - index_x, index_y, index_z (optional)
        - density
        - chi_11, chi_12, chi_13, chi_21, chi_22, chi_23, chi_31, chi_32, chi_33
        - atom_multiplicity
        - point_multiplicity (optional)
    """

    ITEM_CLASS = ChannelPlusMinus
    ATTR_INDEX = None

    def __init__(self, loop_name: str = None, **kwargs) -> NoReturn:
        super(ChannelPlusMinusL, self).__init__()
        self.__dict__["items"] = form_items_by_dictionary(self.ITEM_CLASS, kwargs)
        self.__dict__["loop_name"] = loop_name




"""Classes AtomSiteScat, AtomSiteScatL."""
from typing import NoReturn
import numpy
import matplotlib.pyplot as plt

from cryspy.A_functions_base.function_1_objects import \
    form_items_by_dictionary

from cryspy.B_parent_classes.cl_1_item import ItemN
from cryspy.B_parent_classes.cl_2_loop import LoopN

from cryspy.C_item_loop_classes.cl_1_atom_type_scat import AtomTypeScat, AtomTypeScatL


class AtomSiteScat(ItemN):
    """Description of magnetic structure factor.

    Attributes
    ----------
        - label (mandatory)
        - lande, kappa, type_symbol (optional)
        - atom_type_scat (internal protected attribute)

    Method
    ------
        - calc_form_factor(sthovl, flag_only_orbital=False)
        - load_atom_type_scat_by_symbol(symbol:str)
    """

    ATTR_MANDATORY_NAMES = ("label", )
    ATTR_MANDATORY_TYPES = (str, )
    ATTR_MANDATORY_CIF = ("label", )

    ATTR_OPTIONAL_NAMES = ("type_symbol", "lande", "kappa")
    ATTR_OPTIONAL_TYPES = (str, float, float)
    ATTR_OPTIONAL_CIF = ("type_symbol", "Lande", "kappa")

    ATTR_NAMES = ATTR_MANDATORY_NAMES + ATTR_OPTIONAL_NAMES
    ATTR_TYPES = ATTR_MANDATORY_TYPES + ATTR_OPTIONAL_TYPES
    ATTR_CIF = ATTR_MANDATORY_CIF + ATTR_OPTIONAL_CIF

    ATTR_INT_NAMES = ()
    ATTR_INT_PROTECTED_NAMES = ("atom_type_scat", )

    # parameters considered are refined parameters
    ATTR_REF = ("lande", "kappa")
    ATTR_SIGMA = tuple([f"{_h:}_sigma" for _h in ATTR_REF])
    ATTR_CONSTR_FLAG = tuple([f"{_h:}_constraint" for _h in ATTR_REF])
    ATTR_REF_FLAG = tuple([f"{_h:}_refinement" for _h in ATTR_REF])
    ATTR_CONSTR_MARK = tuple([f"{_h:}_mark" for _h in ATTR_REF])

    # constraints on the parameters
    D_CONSTRAINTS = {"modulation_flag": ["yes", "y", "no", "n"],
                     "refinement_flags_magnetic": ["S", "M", "A", "SM", "SA",
                                                   "MA", "SMA"]}

    # default values for the parameters
    D_DEFAULT = {"lande": 2., "kappa": 1.}
    for key in ATTR_SIGMA:
        D_DEFAULT[key] = 0.
    for key in (ATTR_CONSTR_FLAG + ATTR_REF_FLAG):
        D_DEFAULT[key] = False
    for key in ATTR_CONSTR_MARK:
        D_DEFAULT[key] = ""

    PREFIX = "atom_site_scat"

    def __init__(self, **kwargs) -> NoReturn:
        super(AtomSiteScat, self).__init__()

        # defined for any integer and float parameters
        D_MIN = {"lande": 0, "kappa": 0}

        # defined for ani integer and float parameters
        D_MAX = {}

        self.__dict__["D_MIN"] = D_MIN
        self.__dict__["D_MAX"] = D_MAX
        for key, attr in self.D_DEFAULT.items():
            setattr(self, key, attr)
        for key, attr in kwargs.items():
            setattr(self, key, attr)

    def calc_form_factor(self, sthovl, flag_only_orbital=False):
        """Calculate form factor."""
        if self.is_attribute("atom_type_scat"):
            atom_type_scat = self.atom_type_scat
        else:
            self.load_atom_type_scat_by_symbol(self.type_symbol)
            atom_type_scat = self.atom_type_scat

        form_factor = atom_type_scat.calc_form_factor(
            sthovl, lande=self.lande, kappa=self.kappa,
            flag_only_orbital=flag_only_orbital)
        return form_factor

    def load_atom_type_scat_by_symbol(self, symbol: str):
        """Load details about atom type scattering."""
        self.type_symbol = symbol
        _a_t_s = AtomTypeScat.form_by_symbol(symbol)
        self.__dict__["atom_type_scat"] = _a_t_s

    def plots(self):
        return [self.plot_form_factor()]

    def plot_form_factor(self):
        """Plot magnetic form factor.
        """
        x_min, x_max = 0, 1.5
        sthovl = numpy.linspace(x_min, x_max, 100)
        try:
            res = self.calc_form_factor(sthovl, flag_only_orbital=False)
        except AttributeError:
            return
        label = self.label
        fig, ax = plt.subplots()
        ax.set_xlabel('sin theta / lambda')
        ax.set_xlim(x_min, x_max)
        ax.set_ylabel('Magnetic form factor')
        ax.plot(sthovl, res, label=label) 
        ax.legend(loc='upper right')
        fig.tight_layout()
        return (fig, ax)

    def get_flags_lande(self):
        res = self.lande_refinement
        return res

    def get_flags_kappa(self):
        res = self.kappa_refinement
        return res

class AtomSiteScatL(LoopN):
    """
    Description of magnetic structure factor by Lande factor and kappa.

    Methods
    -------
        - calc_form_factor(sthovl, flag_only_orbital=False)
        - load_atom_type_scat_by_atom_site(atom_site)

    """

    ITEM_CLASS = AtomSiteScat
    ATTR_INDEX = "label"

    def __init__(self, loop_name: str = None, **kwargs) -> NoReturn:
        super(AtomSiteScatL, self).__init__()
        self.__dict__["items"] = form_items_by_dictionary(self.ITEM_CLASS, kwargs)
        self.__dict__["loop_name"] = loop_name

    def calc_form_factor(self, sthovl, flag_only_orbital=False):
        """Calculate form factor."""
        form_factor = [item.calc_form_factor(
            sthovl, flag_only_orbital=flag_only_orbital)
            for item in self.items]
        return numpy.array(list(zip(*form_factor)), dtype=float)

    def load_atom_type_scat_by_atom_site(self, atom_site) -> NoReturn:
        """Load details about atom type scattering."""
        all([item.load_atom_type_scat_by_symbol(
            atom_site[item.label].type_symbol) for item in self.items])
   
    def plots(self):
        return [self.plot_form_factor()]

    def plot_form_factor(self):
        """Plot magnetic form factor.
        """
        x_min, x_max = 0, 1.5
        sthovl = numpy.linspace(x_min, x_max, 100)
        try:
            res = self.calc_form_factor(sthovl, flag_only_orbital=False)
        except AttributeError:
            return
        labels = self.label
        fig, ax = plt.subplots()
        ax.set_title("Magnetic form factor")
        ax.set_xlabel('sin theta / lambda')
        ax.set_xlim(x_min, x_max)
        ax.set_ylabel('Magnetic form factor')
        for ind in range(res.shape[1]):
            ax.plot(sthovl, res[:, ind], label=labels[ind], alpha=0.7) 
        ax.legend(loc='upper right')
        fig.tight_layout()
        return (fig, ax)

    def report(self):
        s_out = ""
        l_ats = []
        for item in self.items:
            try:
                ats = item.atom_type_scat
                l_ats.append(ats)
            except AttributeError:
                pass
        if len(l_ats) != 0:
            obj = AtomTypeScatL()
            obj.items=l_ats
            s_out = str(obj)
        return s_out
    
    def get_flags_lande(self):
        flags_lande = numpy.stack([item.get_flags_lande() for item in self.items], axis=0)
        return flags_lande

    def get_flags_kappa(self):
        flags_kappa = numpy.stack([item.get_flags_kappa() for item in self.items], axis=0)
        return flags_kappa

# s_cont = """
#  loop_
#  _atom_site_scat_label
#  _atom_site_scat_Lande
#  _atom_site_scat_kappa
#   Fe3A    2.00 1.00
#   Fe3B    2.00 1.00(12)
#   """

# obj = AtomSiteScatL.from_cif(s_cont)
# print(obj, end="\n\n")
# print(obj["Fe3B"], end="\n\n")



from typing import Callable
import numpy


def calc_iint(
        beam_polarization: float, 
        flipper_efficiency: float, 
        f_nucl, f_m_perp,
        axis_z, 
        func_extinction: Callable = None,
        flag_beam_polarization: bool = False, 
        flag_flipper_efficiency: bool = False,
        flag_f_nucl: bool = False, 
        flag_f_m_perp: bool = False,
        dict_in_out: dict = None):
    """Calculate integrated intensities.
    Sizes:
        beam_polarization = float or [n_hkl] / the last option is for TOF
        flipper_efficiency = float or [n_hkl] / the last option is for TOF
        f_nucl = [n_hkl]
        f_m_perp = [3, n_hkl]
        axis_Zlab_XYZ = [3] or [3, n_hkl] / the last option is for powder
        func_extinction is function take as input F_square

    for single crystal measurements:
    axis_Zlab_XYZ = [U_31, U_32, U_33]
    """
    if dict_in_out is None:
        flag_dict = False
    else:
        flag_dict = True

    p_u = numpy.atleast_1d(beam_polarization)
    p_d = numpy.atleast_1d(beam_polarization * (2 * flipper_efficiency - 1))
    flag_f_plus_sq = flag_f_nucl or flag_f_m_perp
    flag_f_minus_sq = flag_f_nucl or flag_f_m_perp
    flag_f_m_perp_xy_sq = flag_f_m_perp

    f_n = numpy.atleast_1d(f_nucl)
    if len(f_m_perp .shape) == 1:
        f_m_perp = numpy.expand_dims(f_m_perp , -1)
    if len(axis_z.shape) == 1:
        axis_z = numpy.expand_dims(axis_z, -1)
    
    f_m_perp_sq = numpy.sum(numpy.square(f_m_perp.real), axis=0) + \
        numpy.sum(numpy.square(f_m_perp.imag), axis=0)
    
    f_m_perp_zlab = numpy.sum(f_m_perp * axis_z, axis=0) 

    f_m_perp_zlab_sq =  numpy.square(f_m_perp_zlab.real) + \
        numpy.square(f_m_perp_zlab.imag)


    f_plus = f_n + f_m_perp_zlab
    f_minus = f_n - f_m_perp_zlab
    f_plus_sq = numpy.square(f_plus.real) + numpy.square(f_plus.imag)
    f_minus_sq = numpy.square(f_minus.real) + numpy.square(f_minus.imag)

    f_m_perp_xy_sq = f_m_perp_sq - f_m_perp_zlab_sq


    if func_extinction is None:
        dder_y_plus, dder_y_minus, dder_y_m_perp_xy = {}, {}, {}
        y_plus = numpy.ones_like(f_plus_sq)
        y_minus = numpy.ones_like(f_minus_sq)
        y_m_perp_xy = numpy.ones_like(f_m_perp_xy_sq)
    else:
        y_plus, dder_y_plus = func_extinction(f_plus_sq, flag_f_sq=flag_f_plus_sq)
        y_minus, dder_y_minus = func_extinction(f_minus_sq, flag_f_sq=flag_f_minus_sq)
        y_m_perp_xy, dder_y_m_perp_xy = func_extinction(f_m_perp_xy_sq, flag_f_sq=flag_f_m_perp_xy_sq)

    fmpx_r, fmpx_i = f_m_perp[0].real, f_m_perp[0].imag
    fmpy_r, fmpy_i = f_m_perp[1].real, f_m_perp[1].imag
    fmpz_r, fmpz_i = f_m_perp[2].real, f_m_perp[2].imag

    chiral_term = 2.*(
        axis_z[0] * (fmpy_i * fmpz_r - fmpy_r * fmpz_i) +
        axis_z[1] * (fmpz_i * fmpx_r - fmpz_r * fmpx_i) +
        axis_z[2] * (fmpx_i * fmpy_r - fmpx_r * fmpy_i))
    
    if flag_dict:
        dict_in_out["chiral_term"] = chiral_term

    iint_plus = 0.5 * ((1. + p_u) * y_plus * f_plus_sq +
                       (1. - p_u) * y_minus * f_minus_sq) + \
                y_m_perp_xy * f_m_perp_xy_sq + \
                p_u * chiral_term

    iint_minus = 0.5 * ((1. - p_d) * y_plus * f_plus_sq +
                        (1. + p_d) * y_minus * f_minus_sq) + \
                 y_m_perp_xy * f_m_perp_xy_sq - \
                 p_d * chiral_term

    if flag_dict:
        dict_in_out["iint_plus"] = iint_plus
        dict_in_out["iint_minus"] = iint_minus
        dict_in_out["y_plus"] = y_plus
        dict_in_out["f_plus_sq"] = f_plus_sq
        dict_in_out["y_minus"] = y_minus
        dict_in_out["f_minus_sq"] = f_minus_sq
        dict_in_out["y_m_perp_xy"] = y_m_perp_xy
        dict_in_out["f_m_perp_xy_sq"] = f_m_perp_xy_sq

    dder_plus = {}
    dder_minus = {}
    if flag_beam_polarization:
        dder_plus["beam_polarization"] = 0.5 * (
            y_plus * f_plus_sq - 
            y_minus * f_minus_sq + 
            2. * chiral_term) * \
            numpy.ones_like(beam_polarization)
        dder_minus["beam_polarization"] = 0.5 * (
            -y_plus * f_plus_sq + 
            y_minus * f_minus_sq - 
            2. * chiral_term) * \
            numpy.ones_like(beam_polarization)*(2.*flipper_efficiency-1.)

    if flag_flipper_efficiency:
        dder_minus["flipper_efficiency"] = beam_polarization*(
            -y_plus * f_plus_sq + 
            y_minus * f_minus_sq - 
            2. * chiral_term) * \
            numpy.ones_like(flipper_efficiency)
    
    if flag_f_nucl:
        f_plus_sq_f_n_real = 2. * f_plus.real 
        f_plus_sq_f_n_imag = 2. * f_plus.imag 
        f_minus_sq_f_n_real = 2. * f_minus.real 
        f_minus_sq_f_n_imag = 2. * f_minus.imag 

        y_plus_f_n_real, y_minus_f_n_real = 0, 0 
        y_plus_f_n_imag, y_minus_f_n_imag = 0, 0 
        if "f_sq" in dder_y_plus.keys(): 
            y_plus_f_n_real = dder_y_plus["f_sq"] * f_plus_sq_f_n_real
            y_plus_f_n_imag = dder_y_plus["f_sq"] * f_plus_sq_f_n_imag
        if "f_sq" in dder_y_minus.keys(): 
            y_minus_f_n_real = dder_y_minus["f_sq"] * f_minus_sq_f_n_real
            y_minus_f_n_imag = dder_y_minus["f_sq"] * f_minus_sq_f_n_imag

        dder_plus["f_nucl_real"] =  0.5 * (
            (1.+p_u)*(y_plus*f_plus_sq_f_n_real+y_plus_f_n_real*f_plus_sq) +
            (1.-p_u)*(y_minus*f_minus_sq_f_n_real+y_minus_f_n_real*f_minus_sq))
        
        dder_plus["f_nucl_imag"] =  0.5 * (
            (1.+p_u)*(y_plus*f_plus_sq_f_n_imag+y_plus_f_n_imag*f_plus_sq) +
            (1.-p_u)*(y_minus*f_minus_sq_f_n_imag+y_minus_f_n_imag*f_minus_sq))

        dder_minus["f_nucl_real"] =  0.5 * (
            (1.-p_d)*(y_plus*f_plus_sq_f_n_real+y_plus_f_n_real*f_plus_sq) +
            (1.+p_d)*(y_minus*f_minus_sq_f_n_real+y_minus_f_n_real*f_minus_sq))
        
        dder_minus["f_nucl_imag"] =  0.5 * (
            (1.-p_d)*(y_plus*f_plus_sq_f_n_imag+y_plus_f_n_imag*f_plus_sq) +
            (1.+p_d)*(y_minus*f_minus_sq_f_n_imag+y_minus_f_n_imag*f_minus_sq))


    if flag_f_m_perp:
        f_plus_sq_f_m_perp_x_real = 2. * f_plus.real * axis_z[0] 
        f_plus_sq_f_m_perp_x_imag = 2. * f_plus.imag * axis_z[0] 
        f_plus_sq_f_m_perp_y_real = 2. * f_plus.real * axis_z[1] 
        f_plus_sq_f_m_perp_y_imag = 2. * f_plus.imag * axis_z[1] 
        f_plus_sq_f_m_perp_z_real = 2. * f_plus.real * axis_z[2] 
        f_plus_sq_f_m_perp_z_imag = 2. * f_plus.imag * axis_z[2]
        
        f_minus_sq_f_m_perp_x_real = -2. * f_minus.real * axis_z[0] 
        f_minus_sq_f_m_perp_x_imag = -2. * f_minus.imag * axis_z[0] 
        f_minus_sq_f_m_perp_y_real = -2. * f_minus.real * axis_z[1] 
        f_minus_sq_f_m_perp_y_imag = -2. * f_minus.imag * axis_z[1] 
        f_minus_sq_f_m_perp_z_real = -2. * f_minus.real * axis_z[2] 
        f_minus_sq_f_m_perp_z_imag = -2. * f_minus.imag * axis_z[2] 
        
        zx, zy, zz = axis_z[0], axis_z[1], axis_z[2]
        z_sq = numpy.square(axis_z)
        zx_sq, zy_sq, zz_sq = z_sq[0], z_sq[1], z_sq[2]

        f_m_perp_xy_sq_f_m_perp_x_real = 2. * (
            (1. - zx_sq) * fmpx_r - zx * (zy * fmpy_r + zz * fmpz_r))
        f_m_perp_xy_sq_f_m_perp_x_imag = 2. * (
            (1. - zx_sq) * fmpx_i - zx * (zy * fmpy_i + zz * fmpz_i))
        f_m_perp_xy_sq_f_m_perp_y_real = 2. * (
            (1. - zy_sq) * fmpy_r - zy * (zz * fmpz_r + zx * fmpx_r))
        f_m_perp_xy_sq_f_m_perp_y_imag = 2. * (
            (1. - zy_sq) * fmpy_i - zy * (zz * fmpz_i + zx * fmpx_i))
        f_m_perp_xy_sq_f_m_perp_z_real = 2. * (
            (1. - zz_sq) * fmpz_r - zz * (zx * fmpx_r + zy * fmpy_r))
        f_m_perp_xy_sq_f_m_perp_z_imag = 2. * (
            (1. - zz_sq) * fmpz_i - zz * (zx * fmpx_i + zy * fmpy_i))

        chiral_term_f_m_perp_x_real = 2. * (zy * fmpz_i - zz * fmpy_i)
        chiral_term_f_m_perp_x_imag = -2. * (zy * fmpz_r - zz * fmpy_r)
        chiral_term_f_m_perp_y_real = -2. * (zx * fmpz_i - zz * fmpx_i)
        chiral_term_f_m_perp_y_imag = 2. * (zx * fmpz_r - zz * fmpx_r) 
        chiral_term_f_m_perp_z_real = 2. * (zx * fmpy_i - zy * fmpx_i) 
        chiral_term_f_m_perp_z_imag = -2. * (zx * fmpy_r - zy * fmpx_r)


        y_plus_f_m_perp_x_real, y_plus_f_m_perp_x_imag = 0., 0. 
        y_plus_f_m_perp_y_real, y_plus_f_m_perp_y_imag = 0., 0. 
        y_plus_f_m_perp_z_real, y_plus_f_m_perp_z_imag = 0., 0. 
        y_minus_f_m_perp_x_real, y_minus_f_m_perp_x_imag = 0., 0. 
        y_minus_f_m_perp_y_real, y_minus_f_m_perp_y_imag = 0., 0. 
        y_minus_f_m_perp_z_real, y_minus_f_m_perp_z_imag = 0., 0. 
        y_m_perp_xy_f_m_perp_x_real, y_m_perp_xy_f_m_perp_x_imag = 0., 0. 
        y_m_perp_xy_f_m_perp_y_real, y_m_perp_xy_f_m_perp_y_imag = 0., 0. 
        y_m_perp_xy_f_m_perp_z_real, y_m_perp_xy_f_m_perp_z_imag = 0., 0. 
        if "f_sq" in dder_y_plus.keys(): 
            y_plus_f_m_perp_x_real = dder_y_plus["f_sq"]*f_plus_sq_f_m_perp_x_real
            y_plus_f_m_perp_x_imag = dder_y_plus["f_sq"]*f_plus_sq_f_m_perp_x_imag
            y_plus_f_m_perp_y_real = dder_y_plus["f_sq"]*f_plus_sq_f_m_perp_y_real
            y_plus_f_m_perp_y_imag = dder_y_plus["f_sq"]*f_plus_sq_f_m_perp_y_imag
            y_plus_f_m_perp_z_real = dder_y_plus["f_sq"]*f_plus_sq_f_m_perp_z_real
            y_plus_f_m_perp_z_imag = dder_y_plus["f_sq"]*f_plus_sq_f_m_perp_z_imag
        if "f_sq" in dder_y_minus.keys(): 
            y_minus_f_m_perp_x_real = dder_y_minus["f_sq"]*f_minus_sq_f_m_perp_x_real
            y_minus_f_m_perp_x_imag = dder_y_minus["f_sq"]*f_minus_sq_f_m_perp_x_imag
            y_minus_f_m_perp_y_real = dder_y_minus["f_sq"]*f_minus_sq_f_m_perp_y_real
            y_minus_f_m_perp_y_imag = dder_y_minus["f_sq"]*f_minus_sq_f_m_perp_y_imag
            y_minus_f_m_perp_z_real = dder_y_minus["f_sq"]*f_minus_sq_f_m_perp_z_real
            y_minus_f_m_perp_z_imag = dder_y_minus["f_sq"]*f_minus_sq_f_m_perp_z_imag
        if "f_sq" in dder_y_m_perp_xy.keys(): 
            y_m_perp_xy_f_m_perp_x_real = dder_y_m_perp_xy["f_sq"]*f_m_perp_xy_sq_f_m_perp_x_real
            y_m_perp_xy_f_m_perp_x_imag = dder_y_m_perp_xy["f_sq"]*f_m_perp_xy_sq_f_m_perp_x_imag
            y_m_perp_xy_f_m_perp_y_real = dder_y_m_perp_xy["f_sq"]*f_m_perp_xy_sq_f_m_perp_y_real
            y_m_perp_xy_f_m_perp_y_imag = dder_y_m_perp_xy["f_sq"]*f_m_perp_xy_sq_f_m_perp_y_imag
            y_m_perp_xy_f_m_perp_z_real = dder_y_m_perp_xy["f_sq"]*f_m_perp_xy_sq_f_m_perp_z_real
            y_m_perp_xy_f_m_perp_z_imag = dder_y_m_perp_xy["f_sq"]*f_m_perp_xy_sq_f_m_perp_z_imag


        dder_plus_f_m_perp_x_real =  0.5*(
            (1. + p_u) * (y_plus * f_plus_sq_f_m_perp_x_real + y_plus_f_m_perp_x_real * f_plus_sq) +
            (1. - p_u) * (y_minus * f_minus_sq_f_m_perp_x_real + y_minus_f_m_perp_x_real * f_minus_sq)) + \
            y_m_perp_xy_f_m_perp_x_real * f_m_perp_xy_sq + y_m_perp_xy * f_m_perp_xy_sq_f_m_perp_x_real + \
            p_u * chiral_term_f_m_perp_x_real
        dder_plus_f_m_perp_x_imag =  0.5*(
            (1. + p_u) * (y_plus * f_plus_sq_f_m_perp_x_imag + y_plus_f_m_perp_x_imag * f_plus_sq) +
            (1. - p_u) * (y_minus * f_minus_sq_f_m_perp_x_imag + y_minus_f_m_perp_x_imag * f_minus_sq)) + \
            y_m_perp_xy_f_m_perp_x_imag * f_m_perp_xy_sq + y_m_perp_xy * f_m_perp_xy_sq_f_m_perp_x_imag + \
            p_u * chiral_term_f_m_perp_x_imag

        dder_plus_f_m_perp_y_real =  0.5*(
            (1. + p_u) * (y_plus * f_plus_sq_f_m_perp_y_real + y_plus_f_m_perp_y_real * f_plus_sq) +
            (1. - p_u) * (y_minus * f_minus_sq_f_m_perp_y_real + y_minus_f_m_perp_y_real * f_minus_sq)) + \
            y_m_perp_xy_f_m_perp_y_real * f_m_perp_xy_sq + y_m_perp_xy * f_m_perp_xy_sq_f_m_perp_y_real + \
            p_u * chiral_term_f_m_perp_y_real
        dder_plus_f_m_perp_y_imag =  0.5*(
            (1. + p_u) * (y_plus * f_plus_sq_f_m_perp_y_imag + y_plus_f_m_perp_y_imag * f_plus_sq) +
            (1. - p_u) * (y_minus * f_minus_sq_f_m_perp_y_imag + y_minus_f_m_perp_y_imag * f_minus_sq)) + \
            y_m_perp_xy_f_m_perp_y_imag * f_m_perp_xy_sq + y_m_perp_xy * f_m_perp_xy_sq_f_m_perp_y_imag + \
            p_u * chiral_term_f_m_perp_y_imag

        dder_plus_f_m_perp_z_real =  0.5*(
            (1. + p_u) * (y_plus * f_plus_sq_f_m_perp_z_real + y_plus_f_m_perp_z_real * f_plus_sq) +
            (1. - p_u) * (y_minus * f_minus_sq_f_m_perp_z_real + y_minus_f_m_perp_z_real * f_minus_sq)) + \
            y_m_perp_xy_f_m_perp_z_real * f_m_perp_xy_sq + y_m_perp_xy * f_m_perp_xy_sq_f_m_perp_z_real + \
            p_u * chiral_term_f_m_perp_z_real
        dder_plus_f_m_perp_z_imag =  0.5*(
            (1. + p_u) * (y_plus * f_plus_sq_f_m_perp_z_imag + y_plus_f_m_perp_z_imag * f_plus_sq) +
            (1. - p_u) * (y_minus * f_minus_sq_f_m_perp_z_imag + y_minus_f_m_perp_z_imag * f_minus_sq)) + \
            y_m_perp_xy_f_m_perp_z_imag * f_m_perp_xy_sq + y_m_perp_xy * f_m_perp_xy_sq_f_m_perp_z_imag + \
            p_u * chiral_term_f_m_perp_z_imag


        dder_plus["f_m_perp_real"] = numpy.stack([
            dder_plus_f_m_perp_x_real, 
            dder_plus_f_m_perp_y_real, 
            dder_plus_f_m_perp_z_real],
            axis=0) 

        dder_plus["f_m_perp_imag"] = numpy.stack([
            dder_plus_f_m_perp_x_imag,
            dder_plus_f_m_perp_y_imag, 
            dder_plus_f_m_perp_z_imag],
            axis=0)


        dder_minus_f_m_perp_x_real =  0.5*(
            (1. - p_d) * (y_plus * f_plus_sq_f_m_perp_x_real + y_plus_f_m_perp_x_real * f_plus_sq) +
            (1. + p_d) * (y_minus * f_minus_sq_f_m_perp_x_real + y_minus_f_m_perp_x_real * f_minus_sq)) + \
            y_m_perp_xy_f_m_perp_x_real * f_m_perp_xy_sq + y_m_perp_xy * f_m_perp_xy_sq_f_m_perp_x_real - \
            p_d * chiral_term_f_m_perp_x_real
        dder_minus_f_m_perp_x_imag =  0.5*(
            (1. - p_d) * (y_plus * f_plus_sq_f_m_perp_x_imag + y_plus_f_m_perp_x_imag * f_plus_sq) +
            (1. + p_d) * (y_minus * f_minus_sq_f_m_perp_x_imag + y_minus_f_m_perp_x_imag * f_minus_sq)) + \
            y_m_perp_xy_f_m_perp_x_imag * f_m_perp_xy_sq + y_m_perp_xy * f_m_perp_xy_sq_f_m_perp_x_imag - \
            p_d * chiral_term_f_m_perp_x_imag

        dder_minus_f_m_perp_y_real =  0.5*(
            (1. - p_d) * (y_plus * f_plus_sq_f_m_perp_y_real + y_plus_f_m_perp_y_real * f_plus_sq) +
            (1. + p_d) * (y_minus * f_minus_sq_f_m_perp_y_real + y_minus_f_m_perp_y_real * f_minus_sq)) + \
            y_m_perp_xy_f_m_perp_y_real * f_m_perp_xy_sq + y_m_perp_xy * f_m_perp_xy_sq_f_m_perp_y_real - \
            p_d * chiral_term_f_m_perp_y_real
        dder_minus_f_m_perp_y_imag =  0.5*(
            (1. - p_d) * (y_plus * f_plus_sq_f_m_perp_y_imag + y_plus_f_m_perp_y_imag * f_plus_sq) +
            (1. + p_d) * (y_minus * f_minus_sq_f_m_perp_y_imag + y_minus_f_m_perp_y_imag * f_minus_sq)) + \
            y_m_perp_xy_f_m_perp_y_imag * f_m_perp_xy_sq + y_m_perp_xy * f_m_perp_xy_sq_f_m_perp_y_imag - \
            p_d * chiral_term_f_m_perp_y_imag

        dder_minus_f_m_perp_z_real =  0.5*(
            (1. - p_d) * (y_plus * f_plus_sq_f_m_perp_z_real + y_plus_f_m_perp_z_real * f_plus_sq) +
            (1. + p_d) * (y_minus * f_minus_sq_f_m_perp_z_real + y_minus_f_m_perp_z_real * f_minus_sq)) + \
            y_m_perp_xy_f_m_perp_z_real * f_m_perp_xy_sq + y_m_perp_xy * f_m_perp_xy_sq_f_m_perp_z_real - \
            p_d * chiral_term_f_m_perp_z_real
        dder_minus_f_m_perp_z_imag =  0.5*(
            (1. - p_d) * (y_plus * f_plus_sq_f_m_perp_z_imag + y_plus_f_m_perp_z_imag * f_plus_sq) +
            (1. + p_d) * (y_minus * f_minus_sq_f_m_perp_z_imag + y_minus_f_m_perp_z_imag * f_minus_sq)) + \
            y_m_perp_xy_f_m_perp_z_imag * f_m_perp_xy_sq + y_m_perp_xy * f_m_perp_xy_sq_f_m_perp_z_imag - \
            p_d * chiral_term_f_m_perp_z_imag


        dder_minus["f_m_perp_real"] = numpy.stack([
            dder_minus_f_m_perp_x_real, 
            dder_minus_f_m_perp_y_real, 
            dder_minus_f_m_perp_z_real],
            axis=0) 
        dder_minus["f_m_perp_imag"] = numpy.stack([
            dder_minus_f_m_perp_x_imag, 
            dder_minus_f_m_perp_y_imag, 
            dder_minus_f_m_perp_z_imag],
            axis=0)

    extinction_keys = dder_y_plus.keys()
    if len(extinction_keys) != 0:
        for key in extinction_keys:
            if key == "f_sq":
                pass
            else:
                dder_plus[key] = \
                    0.5*((1.+p_u)*f_plus_sq*dder_y_plus[key] +
                         (1.-p_u)*f_minus_sq*dder_y_minus[key]) + \
                    f_m_perp_xy_sq * dder_y_m_perp_xy[key]
                dder_minus[key] = \
                    0.5*((1.-p_d)*f_plus_sq*dder_y_plus[key] +
                         (1.+p_d)*f_minus_sq*dder_y_minus[key]) + \
                    f_m_perp_xy_sq * dder_y_m_perp_xy[key]
    return iint_plus, iint_minus, dder_plus, dder_minus



def calc_flip_ratio_by_iint(
        iint_plus, iint_minus, c_lambda2: float = None, iint_2hkl = None,
        flag_iint_plus: bool = False, flag_iint_minus: bool = False, 
        flag_c_lambda2: bool = False, flag_iint_2hkl: bool = False):
    """Calculate flip ratio by given integrated intensities.
    """
    dder = {}
    cond = (c_lambda2 is None) or (iint_2hkl is None)
    if cond:
        signal_plus = iint_plus 
        signal_minus = iint_minus 
    else:
        signal_plus = iint_plus + c_lambda2*iint_2hkl
        signal_minus = iint_minus + c_lambda2*iint_2hkl

    flip_ratio = signal_plus/signal_minus

    if flag_iint_plus:
        dder["iint_plus"] = numpy.ones_like(iint_plus) / signal_minus
    if flag_iint_minus:
        dder["iint_minus"] = -1. * flip_ratio * numpy.ones_like(iint_minus)/ signal_minus
    if flag_c_lambda2:
        dder["c_lambda2"] = numpy.ones_like(c_lambda2)*iint_2hkl / signal_minus - \
            flip_ratio * numpy.ones_like(c_lambda2)*iint_2hkl/ signal_minus
    if flag_iint_2hkl:
        dder["iint_2hkl"] = numpy.ones_like(iint_2hkl)*c_lambda2 / signal_minus - \
            flip_ratio * numpy.ones_like(iint_2hkl)*c_lambda2/ signal_minus
    return flip_ratio, dder


def calc_asymmetry_by_iint(
        iint_plus, iint_minus, c_lambda2: float = None, iint_2hkl = None,
        flag_iint_plus: bool = False, flag_iint_minus: bool = False, 
        flag_c_lambda2: bool = False, flag_iint_2hkl: bool = False):
    """Calculate asymmetry (I^+ - I^-)/(I^+ + I^-).
    """
    dder = {}
    cond = (c_lambda2 is None) or (iint_2hkl is None)
    if cond:
        contamination = 0.
    else:
        contamination = 2.*c_lambda2*iint_2hkl

    denom = (iint_plus + iint_minus + contamination)
    asymmetry = (iint_plus - iint_minus) / denom

    if flag_iint_plus:
        dder["iint_plus"] = (1. / denom - (iint_plus - iint_minus) / numpy.square(denom)) * \
            numpy.ones_like(iint_plus)
    if flag_iint_minus:
        dder["iint_minus"] = (-1. / denom - (iint_plus - iint_minus) / numpy.square(denom)) * \
            numpy.ones_like(iint_minus)
    if flag_c_lambda2:
        dder["c_lambda2"] = 2.*iint_2hkl*( - (iint_plus - iint_minus) / numpy.square(denom)) * \
            numpy.ones_like(c_lambda2)
    if flag_iint_2hkl:
        dder["iint_2hkl"] = 2.*c_lambda2*( - (iint_plus - iint_minus) / numpy.square(denom)) * \
            numpy.ones_like(iint_2hkl)
    return asymmetry, dder


def calc_intensities_by_structure_factors(
        beam_polarization: float, flipper_efficiency: float, f_nucl, f_m_perp,
        axis_z, func_extinction: Callable = None,
        c_lambda2: float=None, f_nucl_2hkl=None, f_m_perp_2hkl=None,
        flag_beam_polarization: bool = False, flag_flipper_efficiency: bool = False,
        flag_f_nucl: bool = False, flag_f_m_perp: bool = False,
        flag_c_lambda2: bool = False,
        flag_f_nucl_2hkl: bool = False, flag_f_m_perp_2hkl: bool = False,
        dict_in_out: dict = None):
    """Calculate flip ratio by given structure factors
    """
    if dict_in_out is None:
        flag_dict = False
        dict_in_out_keys = []
    else:
        flag_dict = True
        dict_in_out_keys = dict_in_out.keys()
    
    iint_plus, iint_minus, dder_plus_hkl, dder_minus_hkl = calc_iint(
        beam_polarization, flipper_efficiency, f_nucl, f_m_perp, axis_z,
        func_extinction=func_extinction,
        flag_beam_polarization=flag_beam_polarization, flag_flipper_efficiency=flag_flipper_efficiency,
        flag_f_nucl=flag_f_nucl, flag_f_m_perp=flag_f_m_perp,
        dict_in_out=dict_in_out)

    cond = (c_lambda2 is None) or (f_nucl_2hkl is None) or (f_m_perp_2hkl is None)
    if isinstance(c_lambda2, numpy.ndarray) and not(cond):
        cond = numpy.any(numpy.isnan(c_lambda2))
    if cond:
        c_lambda2 = None
        iint_2hkl = None
        flag_iint_2hkl = False
        dder_2hkl = {}
    else:
        iint_plus_2hkl, iint_minus_2hkl, dder_plus_2hkl, dder_minus_2hkl = calc_iint(
            0, 0, f_nucl_2hkl, f_m_perp_2hkl, axis_z,
            func_extinction=func_extinction,
            flag_beam_polarization=False, flag_flipper_efficiency=False,
            flag_f_nucl=flag_f_nucl_2hkl, flag_f_m_perp=flag_f_m_perp_2hkl,
            dict_in_out=None)
        iint_2hkl = iint_plus_2hkl
        dder_2hkl = dder_plus_2hkl
        flag_iint_2hkl = len(dder_2hkl.keys()) > 0
        
        iint_plus += c_lambda2*iint_2hkl
        iint_minus += c_lambda2*iint_2hkl

        if flag_f_nucl_2hkl:
            dder_2hkl["f_nucl_2hkl_real"] = dder_2hkl.pop("f_nucl_real")
            dder_2hkl["f_nucl_2hkl_imag"] = dder_2hkl.pop("f_nucl_imag")
        if flag_f_m_perp_2hkl:
            dder_2hkl["f_m_perp_2hkl_real"] = dder_2hkl.pop("f_m_perp_real")
            dder_2hkl["f_m_perp_2hkl_imag"] = dder_2hkl.pop("f_m_perp_imag")
    
    dder_plus = {}
    dder_minus = {}
    keys_plus = dder_plus_hkl.keys()
    keys_minus = dder_minus_hkl.keys()
    keys_2h = dder_2hkl.keys()
    set_key = set(keys_plus) | set(keys_minus) | set(keys_2h)
    for key in set_key:
        flag_1 = key in keys_plus
        flag_2 = key in keys_minus
        flag_3 = key in keys_2h
        if flag_1:
            term_1 = dder_plus_hkl[key]
        if flag_2:
            term_2 = dder_minus_hkl[key]
        if flag_3:
            term_3 = dder_2hkl[key]

        if flag_1 and flag_3:
            dder_plus[key] = term_1 + c_lambda2 * term_3
        elif flag_1:
            dder_plus[key] = term_1 
        elif flag_3:
            dder_plus[key] = c_lambda2 * term_3
        
        if flag_2 and flag_3:
            dder_minus[key] = term_2 + c_lambda2 * term_3
        elif flag_2:
            dder_minus[key] = term_2
        elif flag_3:
            dder_minus[key] = c_lambda2 * term_3

    if flag_c_lambda2:
        dder_plus["c_lambda2"] = iint_2hkl
        dder_minus["c_lambda2"] = iint_2hkl
    return iint_plus, iint_minus, dder_plus, dder_minus


def calc_flip_ratio_by_structure_factors(
        beam_polarization: float, flipper_efficiency: float, f_nucl, f_m_perp,
        axis_z, func_extinction: Callable = None,
        c_lambda2: float=None, f_nucl_2hkl=None, f_m_perp_2hkl=None,
        flag_beam_polarization: bool = False,
        flag_flipper_efficiency: bool = False,
        flag_f_nucl: bool = False, flag_f_m_perp: bool = False,
        flag_c_lambda2: bool = False,
        flag_f_nucl_2hkl: bool = False, flag_f_m_perp_2hkl: bool = False,
        dict_in_out: dict = None, flag_use_precalculated_data: bool = False):
    """Calculate flip ratio by given structure factors
    """
    if dict_in_out is None:
        flag_dict = False
        dict_in_out_keys = []
    else:
        flag_dict = True
        dict_in_out_keys = dict_in_out.keys()

    iint_plus, iint_minus, dder_plus, dder_minus = \
        calc_intensities_by_structure_factors(
        beam_polarization, flipper_efficiency, f_nucl, f_m_perp, axis_z,
        func_extinction=func_extinction,
        c_lambda2=c_lambda2, f_nucl_2hkl=f_nucl_2hkl, f_m_perp_2hkl=f_m_perp_2hkl,
        flag_beam_polarization=flag_beam_polarization, flag_flipper_efficiency=flag_flipper_efficiency,
        flag_f_nucl=flag_f_nucl, flag_f_m_perp=flag_f_m_perp,
        flag_c_lambda2=flag_c_lambda2,
        flag_f_nucl_2hkl=flag_f_nucl_2hkl, flag_f_m_perp_2hkl=flag_f_m_perp_2hkl,
        dict_in_out=dict_in_out)

    flip_ration = iint_plus/iint_minus

    dder = {}
    keys_plus = dder_plus.keys()
    keys_minus = dder_minus.keys()
    set_key = set(keys_plus) | set(keys_minus)
    for key in set_key:
        flag_1 = key in keys_plus
        flag_2 = key in keys_minus
        if flag_1 and flag_2:
            dder[key] = dder_plus[key]/iint_minus - flip_ration*dder_minus[key]/iint_minus
        elif flag_1:
            dder[key] = dder_plus[key]/iint_minus
        elif flag_2:
            dder[key] = - flip_ration*dder_minus[key]/iint_minus
    return flip_ration, dder





"""Classes Pd2dPeak, Pd2dPeakL."""
from typing import NoReturn

from cryspy.A_functions_base.function_1_objects import \
    form_items_by_dictionary

from cryspy.B_parent_classes.cl_1_item import ItemN
from cryspy.B_parent_classes.cl_2_loop import LoopN


class Pd2dPeak(ItemN):
    """
    Pd2dPeak class.

    This section contains peak information extracted from the
    measured or, if present, the processed diffractogram.

    Attributes
    ----------
        - index_h, index_k, index_l (mandatory)
        - index_mult, ttheta, f_nucl_sq, f_m_p_sin_sq
        - f_m_p_cos_sq, cross_sin, width_ttheta
    """

    ATTR_MANDATORY_NAMES = ("index_h", "index_k", "index_l")
    ATTR_MANDATORY_TYPES = (int, int, int)
    ATTR_MANDATORY_CIF = ("index_h", "index_k", "index_l")

    ATTR_OPTIONAL_NAMES = ("index_mult", "ttheta", "f_nucl_sq", "f_m_p_sin_sq",
                           "f_m_p_cos_sq", "cross_sin", "width_ttheta",
                           "sintlambda", "d_spacing")
    ATTR_OPTIONAL_TYPES = (int, float, float, float, float, float, float)
    ATTR_OPTIONAL_CIF = ("index_mult", "2theta", "f_nucl_sq", "f_m_p_sin_sq",
                         "f_m_p_cos_sq", "cross_sin", "width_2theta",
                         "sint/lambda", "d_spacing")

    ATTR_NAMES = ATTR_MANDATORY_NAMES + ATTR_OPTIONAL_NAMES
    ATTR_TYPES = ATTR_MANDATORY_TYPES + ATTR_OPTIONAL_TYPES
    ATTR_CIF = ATTR_MANDATORY_CIF + ATTR_OPTIONAL_CIF

    ATTR_INT_NAMES = ()
    ATTR_INT_PROTECTED_NAMES = ()

    # parameters considered are refined parameters
    ATTR_REF = ()
    ATTR_SIGMA = tuple([f"{_h:}_sigma" for _h in ATTR_REF])
    ATTR_CONSTR_FLAG = tuple([f"{_h:}_constraint" for _h in ATTR_REF])
    ATTR_REF_FLAG = tuple([f"{_h:}_refinement" for _h in ATTR_REF])
    ATTR_CONSTR_MARK = tuple([f"{_h:}_mark" for _h in ATTR_REF])

    # formats if cif format
    D_FORMATS = {"ttheta": "{:.2f}", "f_nucl_sq": "{:.2f}",
                 "f_m_p_sin_sq": "{:.2f}", "f_m_p_cos_sq": "{:.2f}",
                 "cross_sin": "{:.2f}", "width_ttheta": "{:.5f}",
                 "sintlambda": "{:.5f}", "d_spacing": "{:.5f}"}

    # constraints on the parameters
    D_CONSTRAINTS = {}

    # default values for the parameters
    D_DEFAULT = {}
    for key in ATTR_SIGMA:
        D_DEFAULT[key] = 0.
    for key in (ATTR_CONSTR_FLAG + ATTR_REF_FLAG):
        D_DEFAULT[key] = False
    for key in ATTR_CONSTR_MARK:
        D_DEFAULT[key] = ""

    PREFIX = "pd2d_peak"

    def __init__(self, **kwargs) -> NoReturn:
        super(Pd2dPeak, self).__init__()

        # defined for any integer and float parameters
        D_MIN = {"ttheta": 0., "width_ttheta": 0.,
                 "sintlambda": 0., "d_spacing": 0.}

        # defined for ani integer and float parameters
        D_MAX = {}

        self.__dict__["D_MIN"] = D_MIN
        self.__dict__["D_MAX"] = D_MAX
        for key, attr in self.D_DEFAULT.items():
            setattr(self, key, attr)
        for key, attr in kwargs.items():
            setattr(self, key, attr)


class Pd2dPeakL(LoopN):
    """Pd2dPeakL class."""

    ITEM_CLASS = Pd2dPeak
    ATTR_INDEX = None

    def __init__(self, loop_name: str = None, **kwargs) -> NoReturn:
        super(Pd2dPeakL, self).__init__()
        self.__dict__["items"] = form_items_by_dictionary(self.ITEM_CLASS, kwargs)
        self.__dict__["loop_name"] = loop_name


# s_cont = """
#  loop_
#  _pd2d_peak_index_h
#  _pd2d_peak_index_k
#  _pd2d_peak_index_l
#  _pd2d_peak_index_mult
#  _pd2d_peak_ttheta
#  _pd2d_peak_f_nucl_sq
#  _pd2d_peak_f_m_p_sin_sq
#  _pd2d_peak_f_m_p_cos_sq
#  _pd2d_peak_cross_sin
#  _pd2d_peak_width_ttheta
#   2  2  0  4 17.2 100.0 101.2   90.0  87.4 2.3
#   2  1  3  2 19.1  78.6 101.0   92.0  82.7 5.7
# """

# obj = Pd2dPeakL.from_cif(s_cont)
# print(obj, end="\n\n")



import numpy
from typing import NoReturn

from cryspy.A_functions_base.function_1_objects import \
    form_items_by_dictionary

from cryspy.B_parent_classes.cl_1_item import ItemN
from cryspy.B_parent_classes.cl_2_loop import LoopN

class AtomType(ItemN):
    """Details about properties of the atom.

    Data items in the ATOM_TYPE category record details about
    properties of the atoms that occupy the atom sites, such as the
    atomic scattering factors.

    Mandatory attributes:
        - label
        - type_symbol
        - fract_x
        - fract_y
        - fract_z

    Optional attributes:
        - occupancy
        - adp_type
        - u_iso_or_equiv
        - u_equiv_geom_mean
        - b_iso_or_equiv
        - multiplicity
        - wyckoff_symbol
        - cartn_x
        - cartn_y
        - cartn_z

    Internal attributes:
        - scat_length_neutron

    Internal protected attributes:
        - space_group_wyckoff
        - constr_number
    """
    ATTR_MANDATORY_NAMES = ("symbol", )
    ATTR_MANDATORY_TYPES = (str, )
    ATTR_MANDATORY_CIF = ("symbol", )

    ATTR_OPTIONAL_NAMES = (
        "analytical_mass", "description", "number_in_cell", "oxidation_number",
        "radius_bond", "radius_contact", "cromer_mann_a1", "cromer_mann_a2",
        "cromer_mann_a3", "cromer_mann_a4", "cromer_mann_b1", "cromer_mann_b2",
        "cromer_mann_b3", "cromer_mann_b4", "cromer_mann_c",
        "scat_dispersion_real", "scat_dispersion_imag", "dispersion_source",
        "scat_length_neutron", "scat_source", "scat_versus_stol_list")
    ATTR_OPTIONAL_TYPES = (float, str, int, float, float, float, float, float,
                           float, float, float, float, float, float, float,
                           float, float, str, complex, str, float)
    ATTR_OPTIONAL_CIF = (
        "analytical_mass", "description", "number_in_cell", "oxidation_number",
        "radius_bond", "radius_contact", "cromer_mann_a1", "cromer_mann_a2",
        "cromer_mann_a3", "cromer_mann_a4", "cromer_mann_b1", "cromer_mann_b2",
        "cromer_mann_b3", "cromer_mann_b4", "cromer_mann_c",
        "scat_dispersion_real", "scat_dispersion_imag", "dispersion_source",
        "scat_length_neutron", "scat_source", "scat_versus_stol_list")

    ATTR_NAMES = ATTR_MANDATORY_NAMES + ATTR_OPTIONAL_NAMES
    ATTR_TYPES = ATTR_MANDATORY_TYPES + ATTR_OPTIONAL_TYPES
    ATTR_CIF = ATTR_MANDATORY_CIF + ATTR_OPTIONAL_CIF

    ATTR_INT_NAMES = ()
    ATTR_INT_PROTECTED_NAMES = ()

    # parameters considered are refined parameters
    ATTR_REF = ()
    ATTR_SIGMA = tuple([f"{_h:}_sigma" for _h in ATTR_REF])
    ATTR_CONSTR_FLAG = tuple([f"{_h:}_constraint" for _h in ATTR_REF])
    ATTR_REF_FLAG = tuple([f"{_h:}_refinement" for _h in ATTR_REF])
    ATTR_CONSTR_MARK = tuple([f"{_h:}_mark" for _h in ATTR_REF])

    # constraints on the parameters
    D_CONSTRAINTS = {}

    # default values for the parameters
    D_DEFAULT = {}
    for key in ATTR_SIGMA:
        D_DEFAULT[key] = 0.
    for key in (ATTR_CONSTR_FLAG + ATTR_REF_FLAG):
        D_DEFAULT[key] = False
    for key in ATTR_CONSTR_MARK:
        D_DEFAULT[key] = ""

    PREFIX = "atom_type"

    def __init__(self, **kwargs) -> NoReturn:
        super(AtomType, self).__init__()

        # defined for any integer and float parameters
        D_MIN = {}

        # defined for ani integer and float parameters
        D_MAX = {}

        self.__dict__["D_MIN"] = D_MIN
        self.__dict__["D_MAX"] = D_MAX
        for key, attr in self.D_DEFAULT.items():
            setattr(self, key, attr)
        for key, attr in kwargs.items():
            setattr(self, key, attr)



class AtomTypeL(LoopN):
    """Details about properties of the atoms.
    """
    ITEM_CLASS = AtomType
    ATTR_INDEX = "symbol"
    def __init__(self, loop_name: str = None, **kwargs) -> NoReturn:
        super(AtomTypeL, self).__init__()
        self.__dict__["items"] = form_items_by_dictionary(self.ITEM_CLASS, kwargs)
        self.__dict__["loop_name"] = loop_name



# s_cont = """
#  loop_
#  _atom_type_symbol
#  _atom_type_oxidation_number
#  _atom_type_number_in_cell
#  _atom_type_scat_dispersion_real
#  _atom_type_scat_dispersion_imag
#  _atom_type_scat_source
#    C 0 72  .017  .009  International_Tables_Vol_IV_Table_2.2B
#    H 0 100  0     0    International_Tables_Vol_IV_Table_2.2B
#    O 0 12  .047  .032  International_Tables_Vol_IV_Table_2.2B
#    N 0 4   .029  .018  International_Tables_Vol_IV_Table_2.2B
#   """

# obj = AtomTypeL.from_cif(s_cont)
# print(obj, end="\n\n")
# print(obj["O"].scat_length_neutron, end="\n\n")
